"""``photolossrates`` - Determine photoionization and photodissociation rates"""
import os
import pandas as pd
import astropy.units as u
from nexoclom import __file__ as basefile


class PhotoRate:
    r"""Determine photoreactions and photorates for a species.

    **Parameters**

    species
        Species to compute rates for.

    aplanet
        Distance from the Sun. Default is 1 AU. Given as either a numeric
        type or an astropy quantity with length units.

    **Class Attributes**

    species
        Species

    aplanet
        Distance from the Sun; astropy quantity with units AU

    rate
        Reaction rate; astropy quantity with units s^{-1}. Rate is the sum
        of all possible reactions for the species.

    reactions
        Pandas dataframe with columns for reaction and rate (in s^{-1}) for
        each reaction for the species. This can be used to determine the
        products produced by photolysis and photoionization.
        
    **Example**
    ::
        >>> from nexoclom.atomicdata import PhotoRate
        >>> kappa = PhotoRate('Na', 0.33)
        >>> print(kappa)
        Species = Na
        Distance = 0.33 AU
        Rate = 6.666666666666666e-05 1 / s
        >>> print(kappa.rate)
        6.666666666666666e-05 1 / s
        >>> print(kappa.reactions)
                       reaction                  kappa
        0  Na, photon -> Na+, e  6.666666666666666e-05
        >>> kappa = PhotoRate('H_2O')
        >>> print(kappa)
        Species = H_2O
        Distance = 1.0 AU
        Rate = 1.2056349999999999e-05 1 / s
        >>> print(kappa.reactions)
                             reaction     kappa
        0      H_2O, photon -> H_2, O  5.97e-07
        1       H_2O, photon -> OH, H  1.03e-05
        2     H_2O, photon -> H, H, O  7.54e-07
        3   H_2O, photon -> H, OH+, e  5.54e-08
        4   H_2O, photon -> OH, H+, e  1.31e-08
        5    H_2O, photon -> H_2O+, e  3.31e-07
        6  H_2O, photon -> H_2, O+, e  5.85e-09

    """
    def __init__(self, species, aplanet_=1.*u.AU):
        photorates = pd.read_pickle(self.photorates_filename())
        prates = photorates[photorates.species == species]

        if isinstance(aplanet_, type(1*u.au)):
            aplanet = aplanet_.value
        else:
            aplanet = aplanet_

        self.species = species
        self.aplanet = aplanet*u.au

        # Photo rate adjusted to proper heliocentric distance
        if len(prates) == 0:
            print('No photoreactions found')
            self.reactions = None
            self.rate = 1e-30/u.s
        else:
            rates = prates['kappa'].apply(lambda k: k/aplanet**2).values
            self.reactions = prates
            self.rate = rates.sum()/u.s

    def __str__(self):
        output = (f'Species = {self.species}\n'
                  f'Distance = {self.aplanet}\n'
                  f'Rate = {self.rate}')
        return output

    def photorates_filename(self):
        basepath = os.path.dirname(basefile)
        return os.path.join(basepath, 'data', 'Loss', 'photorates.pkl')
        



"""Computes acceleration and ionization on a packet due to specified forces

Gravitational acceleration

Equations of motion:
    dvxdt = sum_objects (GM * (x-x_obj))/(r_obj)^3
    dvydt = sum_objects (GM * (y-y_obj))/(r_obj)^3
    dvzdt = sum_objects (GM * (z-z_obj))/(r_obj)^3
        -- r_obj = sqrt( (x-x_obj)^2 + (y-y_obj)^2 + (z-z_obj)^2 )
    dndt = instantaneous change in density

Current version: Assumes there is only a planet -- does not do moons yet
"""
import numpy as np


def state(x, output):
    # compute gravitational acceleration
    if output.inputs.forces.gravity:
        r3 = (np.linalg.norm(x[:,1:4], axis=1))**3
        agrav = output.GM * x[:,1:4]/r3[:,np.newaxis]
    else:
        agrav = np.zeros(x[:,1:4].shape)
        
    # compute radiation acceleration
    arad = np.zeros(x[:,1:4].shape)
    if output.inputs.forces.radpres:
        rho = np.linalg.norm(x[:,[1,3]], axis=1)
        out_of_shadow = (rho > 1) | (x[:,2] < 0)

        # radial velocity of each packet realtive to the Sun
        vv = x[:,5] + output.vrplanet

        # Compute radiation acceleration
        arad[:,1] = np.interp(vv, output.radpres.velocity,
                              output.radpres.accel) * out_of_shadow
    else:
        pass
    
    # Compute total acceleration
    accel = agrav + arad

    # Compute ionization rate
    if output.inputs.options.lifetime > 0:
        # Explicitly set lifetime
        ionizerate = np.ones(x[:,0].shape)/output.inputs.options.lifetime.value
    else:
        if output.loss_info.photo is not None:
            # Compute photoionization rate
            rho = np.linalg.norm(x[:,[1,3]], axis=1)
            out_of_shadow = (rho > 1) | (x[:,2] < 0)
            photorate = output.loss_info.photo * out_of_shadow
        else:
            photorate = np.zeros(x[:,0].shape)
         
        '''
        magcoord = xyz_to_magcoord(t, x, output.inputs, output.planet)

        if output.loss_info.eimp:
            # Compute electron impact rate
            assert 0, 'Electron impacts not set up yet'
        else:
            eimprate = 0.

        if output.loss_info.chX:
            # Compute charge exchange rate
            assert 0, 'Charge exchange not set up yet'
        else:
            chxrate = 0.
        '''

        ionizerate = photorate  # + eimprate + chxrate

    return accel, ionizerate



import numpy as np
import astropy.units as u
from nexoclom.math import interpu
import pytest


def fn(x):
    return (x**np.random.random()*3 +
            (10*np.random.random()-5)*x**np.random.random()*3 +
            (10*np.random.random()-5))

@pytest.mark.math
def test_interpu():
    xp = np.arange(20)
    fp = fn(xp)
    
    points = np.random.rand(30)*20-5
    result = np.interp(points, xp, fp)
    test = interpu(points*u.km, xp*u.km, fp*u.km**3)

    assert test.value == pytest.approx(result)
    assert test.unit == u.km**3
    
    test = interpu((points*u.km).to(u.imperial.mi), xp*u.km, fp*u.km**3)
    assert test.value == pytest.approx(result)

    with pytest.raises(TypeError):
        interpu(points, xp*u.km, fp*u.km**3)

    with pytest.raises(TypeError):
        interpu(points*u.km, xp, fp*u.km**3)

    with pytest.raises(TypeError):
        interpu(points*u.km, xp*u.km, fp)

    with pytest.raises(u.UnitConversionError):
        interpu(points*u.s, xp*u.km, fp*u.km**3)



import os
import pickle
import astropy.units as u
from nexoclom.atomicdata import gValue, RadPresConst
from pytest import approx
import pytest
# pylint: disable=no-member


with open(os.path.join(os.path.dirname(__file__),
                       'g_value_test_data.pkl'), 'rb') as f:
    gvalue_test_data, radpres_test_data = pickle.load(f)
 
args_gval = [('Na', 5891, 1.5, gvalue_test_data[0]),
             ('Ca', 4227*u.AA, 0.3, gvalue_test_data[1]),
             ('X', 3333*u.AA, 1.0*u.au, gvalue_test_data[2])]
args_rp = [('Na', 1.5, radpres_test_data[0]),
           ('Ca', 0.3*u.au, radpres_test_data[1]),
           ('X', 1.0*u.au, radpres_test_data[2])]

@pytest.mark.atomicdata
@pytest.mark.parametrize('species, wavelength, aplanet, result', args_gval)
def test_gValue(species, wavelength, aplanet, result):
    g = gValue(species, wavelength, aplanet)
    assert g.species == result['species'], 'Species Failure'
    assert g.wavelength == result['wavelength'], 'Wavelength failure'
    assert g.aplanet == result['aplanet'], 'aplanet failure'
    assert g.velocity.value == approx(result['velocity'].value), 'velocity failure'
    assert g.g.value == approx(result['g'].value), 'g-values failure'

@pytest.mark.atomicdata
def test_gvalue_bad_input():
    with pytest.raises(ValueError):
        gValue('Fk', 9999, 1)

@pytest.mark.atomicdata
@pytest.mark.parametrize('species, aplanet, result', args_rp)
def test_radpresconst(species, aplanet, result):
    rp_const = RadPresConst(species, aplanet)
    assert rp_const.species == result['species'], 'Species Failure'
    assert rp_const.aplanet == result['aplanet'], 'aplanet Failure'
    assert rp_const.velocity.value == approx(result['velocity'].value), 'velocity failure'
    assert rp_const.accel.value == approx(result['accel'].value), 'accel failure'

@pytest.mark.atomicdata
def test_radpres_bad_input():
    with pytest.raises(ValueError):
        RadPresConst('Fk', 1)


''' Determine the relative position of object to reference

This is basically a rewrite of cspice_spkezr without having to remember
all the options to set

Inputs:
    obj: target object name
    reference: reference object name
    et: either a single astropy Time object, a single et, a list of Time objects,
        or a numpy array of ETs.

Output:
    pos = position in km
    vel = velocity in km/s
'''
import numpy as np
import spiceypy as spice
from astropy.time import Time
import astropy.units as u
from nexoclom.solarsystem.load_kernels import load_kernels


def relative_position(obj, reference, ettemp, frame='J2000'):
    kernels = load_kernels()

    if isinstance(ettemp, Time):
        et = spice.str2et(ettemp.isot)
    elif isinstance(ettemp, float):
        et = ettemp
    else:
        et = [spice.str2et(e.isot) if type(e) == type(Time.now()) else e
              for e in ettemp]

    abcor = 'LT+S'

    if isinstance(et, float):
        posvel, lt = spice.spkezr(obj, et, frame, abcor, reference)
        pos, vel = posvel[:3], posvel[3:]
    else:
        pos = np.zeros((len(et),3))
        vel = np.zeros((len(et),3))
        for i, t in enumerate(et):
            posvel, lt = spice.spkezr(obj, t, frame, abcor, reference)
            pos[i,:] = np.array(posvel[:3])
            vel[i,:] = np.array(posvel[3:])
    pos *= u.km
    vel *= u.km/u.s

    for k in kernels:
        spice.unload(k)

    return pos, vel



""" Tests Input.__init__() and classes in input_classes.py
Note that this is more of a regression test than a unit test.
Compares with previously computed results."""
import os
import numpy as np
import pytest
import astropy.units as u
from astropy.time import Time
from nexoclom.solarsystem import SSObject
from nexoclom import Input, __file__ as basefile

basepath = os.path.dirname(basefile)
if __name__ == '__main__':
    inputpath = os.path.join('test_data', 'inputfiles')
else:
    inputpath = os.path.join(basepath, 'modelcode', 'tests', 'test_data', 'inputfiles')


@pytest.mark.modelcode
def test_geometry():
    inputfile01 = os.path.join(inputpath, 'Geometry.01.input')
    geometry01 = Input(inputfile01).geometry
    result = {'planet': SSObject('Jupiter'),
              'startpoint': 'Io',
              'objects': {SSObject('Jupiter'), SSObject('Io'), SSObject('Europa')},
              'type': 'geometry without starttime',
              'phi': (1*u.rad, 2*u.rad),
              'subsolarpoint': (3.14*u.rad, 0*u.rad),
              'taa': 1.57*u.rad}
    assert geometry01.__dict__ == result
    
    inputfile02 = os.path.join(inputpath, 'Geometry.02.input')
    geometry02 = Input(inputfile02).geometry
    result = {'planet': SSObject('Jupiter'),
              'startpoint': 'Io',
              'objects': {SSObject('Jupiter'), SSObject('Io')},
              'type': 'geometry with starttime',
              'time': Time('2022-03-08T19:53:21')}
    assert geometry02.__dict__ == result

    inputfile03 = os.path.join(inputpath, 'Geometry.03.input')
    geometry03 = Input(inputfile03).geometry
    result = {'planet':SSObject('Mercury'),
              'startpoint': 'Mercury',
              'objects': {SSObject('Mercury')},
              'type': 'geometry without starttime',
              'subsolarpoint': (0 * u.rad, 0 * u.rad),
              'phi': None,
              'taa': 3.14 * u.rad}
    assert geometry03.__dict__ == result
    
    assert geometry01 == geometry01
    assert geometry01 != geometry02
    assert geometry01 != geometry03

@pytest.mark.modelcode
def test_SurfaceInteraction():
    # sticktype = 'constant'
    inputfile01 = os.path.join(inputpath, 'SurfaceInteraction.01.input')
    interaction01 = Input(inputfile01).surfaceinteraction
    result = {'sticktype': 'constant',
              'stickcoef': 1.,
              'accomfactor': None}
    assert interaction01.__dict__ == result

    inputfile02 = os.path.join(inputpath, 'SurfaceInteraction.02.input')
    interaction02 = Input(inputfile02).surfaceinteraction
    result = {'sticktype': 'constant',
              'stickcoef': 0.5,
              'accomfactor': 0.2}
    assert interaction02.__dict__ == result

    assert interaction01 == interaction01
    assert interaction01 != interaction02

    # sticktype = 'temperature dependent
    inputfile03 = os.path.join(inputpath, 'SurfaceInteraction.03.input')
    interaction03 = Input(inputfile03).surfaceinteraction
    result = {'sticktype': 'temperature dependent',
              'accomfactor': 0.2,
              'A': (1.57014, -0.006262, 0.1614157)}
    assert interaction03.__dict__ == result

    inputfile04 = os.path.join(inputpath, 'SurfaceInteraction.04.input')
    interaction04 = Input(inputfile04).surfaceinteraction
    result = {'sticktype':'temperature dependent',
              'accomfactor':0.5,
              'A': (1., 0.001, 0.2)}
    assert interaction04.__dict__ == result

    inputfile05 = os.path.join(inputpath, 'SurfaceInteraction.05.input')
    interaction05 = Input(inputfile05).surfaceinteraction
    result = {'sticktype':'surface map',
              'stick_mapfile': 'default',
              'coordinate_system': 'solar-fixed',
              'subsolarlon': None,
              'accomfactor':0.5}
    assert interaction05.__dict__ == result

    inputfile06 = os.path.join(inputpath, 'SurfaceInteraction.06.input')
    interaction06 = Input(inputfile06).surfaceinteraction
    result = {'sticktype':'surface map',
              'stick_mapfile': 'Orbit3576.Ca.pkl',
              'coordinate_system': 'solar-fixed',
              'subsolarlon': None,
              'accomfactor':0.5}
    assert interaction06.__dict__ == result

@pytest.mark.modelcode
def test_Forces():
    inputfile01 = os.path.join(inputpath, 'Forces.01.input')
    forces01 = Input(inputfile01).forces
    result = {'gravity': True,
              'radpres': True}
    assert forces01.__dict__ == result

    inputfile02 = os.path.join(inputpath, 'Forces.02.input')
    forces02 = Input(inputfile02).forces
    result = {'gravity': False,
              'radpres': True}
    assert forces02.__dict__ == result

    inputfile03 = os.path.join(inputpath, 'Forces.03.input')
    forces03 = Input(inputfile03).forces
    result = {'gravity': True,
              'radpres': False}
    assert forces03.__dict__ == result

@pytest.mark.modelcode
def test_SpatialDist():
    inputfile01 = os.path.join(inputpath, 'Spatial.01.input')
    spatial01 = Input(inputfile01).spatialdist
    result = {'type': 'uniform',
              'longitude': (0*u.rad, 2*np.pi*u.rad),
              'latitude': (-np.pi/2*u.rad, np.pi/2*u.rad),
              'exobase': 1.}
    assert spatial01.__dict__ == pytest.approx(result)

    inputfile02 = os.path.join(inputpath, 'Spatial.02.input')
    spatial02 = Input(inputfile02).spatialdist
    result = {'type':'uniform',
              'longitude':(0*u.rad, 3.14*u.rad),
              'latitude':(0*u.rad, 0.79*u.rad),
              'exobase':2.1}
    assert spatial02.__dict__ == pytest.approx(result)
    
if __name__ == '__main__':
    # test_geometry()
    # test_SurfaceInteraction()
    # test_Forces()
    test_SpatialDist()



import pytest
import astropy.units as u
from nexoclom.atomicdata import PhotoRate
from pytest import approx


args = [('Na', 1.5, 3.2266666666666665e-06),
        ('Ca', 0.3*u.au, 0.0007777777777777777),
        ('X', 1.0*u.au, 1e-30)]

@pytest.mark.atomicdata
@pytest.mark.parametrize('species, aplanet, result', args)
def test_PhotoRate(species, aplanet, result, capsys):
    rate = PhotoRate(species, aplanet)
    assert rate.species == species
    if isinstance(aplanet, type(1*u.au)):
        assert rate.aplanet.value == aplanet.value
    else:
        assert rate.aplanet.value == aplanet
    assert rate.rate.value == approx(result)
    assert rate.rate.unit == 1./u.s
    _ = capsys.readouterr()

    print(rate)
    captured = capsys.readouterr()
    assert captured[0] == (f'Species = {rate.species}\n'
                           f'Distance = {rate.aplanet}\n'
                           f'Rate = {rate.rate}\n')
    assert captured[1] == ''

    
    


# -*- coding: utf-8 -*-
#
# STSCI documentation build configuration file, created by
# sphinx-quickstart on Thu Oct 22 17:25:41 2015.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import datetime
import importlib
import sys
import os
import sphinx
import stsci_rtd_theme

def setup(app):
    app.add_stylesheet("stsci.css")

# MagicMock takes care of packages that can't be imported on ReadTheDocs
from unittest.mock import MagicMock
class Mock(MagicMock):
    @classmethod
    def __getattr__(cls, name):
        return MagicMock()

MOCK_MODULES = ['psycopg2', 'MESSENGERuvvs', 'solarsystemMB',
                'atomicdataMB', 'nexoclom.state']
sys.modules.update((mod_name, Mock()) for mod_name in MOCK_MODULES)


from distutils.version import LooseVersion
try:
    from ConfigParser import ConfigParser
except ImportError:
    from configparser import ConfigParser
conf = ConfigParser()

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.abspath('../'))
sys.path.insert(0, os.path.abspath('./'))
sys.path.insert(0, os.path.abspath('exts/'))

# -- General configuration ------------------------------------------------
conf.read([os.path.join(os.path.dirname(__file__), '..', 'setup.cfg')])
setup_cfg = dict(conf.items('metadata'))

# If your documentation needs a minimal Sphinx version, state it here.
needs_sphinx = '1.3'

on_rtd = os.environ.get('READTHEDOCS', None) == 'True'


def check_sphinx_version(expected_version):
    sphinx_version = LooseVersion(sphinx.__version__)
    expected_version = LooseVersion(expected_version)
    if sphinx_version < expected_version:
        raise RuntimeError(
            "At least Sphinx version {0} is required to build this "
            "documentation.  Found {1}.".format(
                expected_version, sphinx_version))

# Configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    'python': ('http://docs.python.org/3/', None),
    'numpy': ('http://docs.scipy.org/doc/numpy/', None),
    'scipy': ('http://docs.scipy.org/doc/scipy/reference/', None),
    'matplotlib': ('http://matplotlib.org/', None),
    }

if sys.version_info[0] == 2:
    intersphinx_mapping['python'] = ('http://docs.python.org/2/', None)
    intersphinx_mapping['pythonloc'] = (
        'http://docs.python.org/',
        os.path.abspath(os.path.join(os.path.dirname(__file__),
                                     'local/python2_local_links.inv')))

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.inheritance_diagram',
    'sphinx.ext.viewcode',
    'sphinx.ext.autosummary',
    ]

if on_rtd:
    extensions.append('sphinx.ext.mathjax')
elif LooseVersion(sphinx.__version__) < LooseVersion('1.4'):
    extensions.append('sphinx.ext.pngmath')
else:
    extensions.append('sphinx.ext.imgmath')


# Add any paths that contain templates here, relative to this directory.
# templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# A list of warning types to suppress arbitrary warning messages. We mean to
# override directives in astropy_helpers.sphinx.ext.autodoc_enhancements,
# thus need to ignore those warning. This can be removed once the patch gets
# released in upstream Sphinx (https://github.com/sphinx-doc/sphinx/pull/1843).
# Suppress the warnings requires Sphinx v1.4.2
suppress_warnings = ['app.add_directive', ]


# General information about the project
project = setup_cfg['package_name']
author = setup_cfg['author']
copyright = '{0}, {1}'.format(datetime.datetime.now().year, author)

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# build documents.
#
# The short X.Y version.
package = importlib.import_module(setup_cfg['package_name'])
version = package.__version__.split('-', 1)[0]
# The full version, including alpha/beta/rc tags.
release = package.__version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build']

# The reST default role (used for this markup: `text`) to use for all
# documents.
default_role = 'obj'


# Don't show summaries of the members in each class along with the
# class' docstring
numpydoc_show_class_members = False

autosummary_generate = True

automodapi_toctreedirnm = 'api'

# Class documentation should contain *both* the class docstring and
# the __init__ docstring
autoclass_content = "both"

# Render inheritance diagrams in SVG
graphviz_output_format = "svg"

graphviz_dot_args = [
    '-Nfontsize=10',
    '-Nfontname=Helvetica Neue, Helvetica, Arial, sans-serif',
    '-Efontsize=10',
    '-Efontname=Helvetica Neue, Helvetica, Arial, sans-serif',
    '-Gfontsize=10',
    '-Gfontname=Helvetica Neue, Helvetica, Arial, sans-serif'
]


# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
# keep_warnings = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'stsci_rtd_theme'
html_theme_path = [stsci_rtd_theme.get_html_theme_path()]

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = None

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
# html_extra_path = []

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
html_sidebars = {'**': ['globaltoc.html', 'relations.html', 'searchbox.html']}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
html_domain_indices = True

# If false, no index is generated.
html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = 'nexoclomdoc'


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    'pointsize': '14pt',
    # Additional stuff for the LaTeX preamble.
    'preamble': r'''\usepackage{enumitem} \setlistdepth{99}'''
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
# latex_documents = [
#   ('index', 'nexoclom.tex', u'nexoclom Documentation',
#    u'nexoclom', 'manual'),
# ]

# The name of an image file (relative to this directory) to place at the top of
# the title page.

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# If true, show page references after internal links.
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
latex_show_urls = 'True'

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
latex_domain_indices = True


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ('index', 'nexoclom', u'nexoclom Documentation',
     [u'nexoclom'], 1)
]

# If true, show URL addresses after external links.
man_show_urls = True


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
  ('index', 'nexoclom', u'nexoclom Documentation',
   u'nexoclom', 'nexoclom', 'nexoclom',
   'Miscellaneous'),
]

# Documents to append as an appendix to all manuals.
# texinfo_appendices = []

# If false, no module index is generated.
texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
texinfo_show_urls = 'inline'

# If true, do not generate a @detailmenu in the "Top" node's menu.
# texinfo_no_detailmenu = False


# -- Options for Epub output ----------------------------------------------

# Bibliographic Dublin Core info.
epub_title = u'nexoclom'
epub_author = u'Matthew Burger'
#epub_publisher = u'STSCI'
#epub_copyright = u'2016 STScI'

# The basename for the epub file. It defaults to the project name.
# epub_basename = u'wfc3tools'

# The HTML theme for the epub output. Since the default themes are not
# optimized for small screen space, using the same theme for HTML and
# epub output is usually not wise. This defaults to 'epub', a theme designed
# to save visual space.
epub_theme = 'epub'

# The language of the text. It defaults to the language option
# or en if the language is not set.
# epub_language = ''

# The scheme of the identifier. Typical schemes are ISBN or URL.
# epub_scheme = ''

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
# epub_identifier = ''

# A unique identification for the text.
# epub_uid = ''

# A tuple containing the cover image and cover page html template filenames.
# epub_cover = ()

# A sequence of (type, uri, title) tuples for the guide element of content.opf.
# epub_guide = ()

# HTML files that should be inserted before the pages created by sphinx.
# The format is a list of tuples containing the path and title.
# epub_pre_files = []

# HTML files shat should be inserted after the pages created by sphinx.
# The format is a list of tuples containing the path and title.
# epub_post_files = []

# A list of files that should not be packed into the epub file.
epub_exclude_files = ['search.html']

# The depth of the table of contents in toc.ncx.
# epub_tocdepth = 3

# Allow duplicate toc entries.
# epub_tocdup = True

# Choose between 'default' and 'includehidden'.
# epub_tocscope = 'default'

# Fix unsupported image types using the PIL.
# epub_fix_images = False

# Scale large images.
# epub_max_image_width = 0

# How to display URL addresses: 'footnote', 'no', or 'inline'.
# epub_show_urls = 'inline'

# If false, no index is generated.
# epub_use_index = True



"""Read the model inputs from a file and create the Input object.

The Input object is build from smaller objects defining different model
options.

Geometry
    Defines the Solar System geometry for the Input.

SurfaceInteraction
    Defines the surface interactions.

Forces
    Set which forces act on model particles.

SpatialDist
    Define the initial spatial distribution of particles.

SpeedDist
    Define the initial speed distribution of particles.

AngularDist
    Define the initial angular distribtuion of particles.

Options
    Configure other model parameters
"""
import os
import os.path
import numpy as np
import pandas as pd
from astropy.time import Time
import sqlalchemy as sqla
import sqlalchemy.dialects.postgresql as pg
from nexoclom.modelcode.Output import Output
from nexoclom.utilities import NexoclomConfig
from nexoclom.modelcode.input_classes import (Geometry, SurfaceInteraction,
                                             Forces, SpatialDist, SpeedDist,
                                             AngularDist, Options)
from nexoclom.modelcode.ModelImage import ModelImage

class Input:
    def __init__(self, infile):
        """Read the input options from a file.

        **Parameters**
        
        infile
            Plain text file containing model input parameters. See
            :doc:`inputfiles` for a description of the input file format.

        **Class Attributes**

        * geometry
        
        * surface_interaction
        
        * forces
        
        * spatialdist
        
        * speeddist
        
        * angulardist
        
        * options
        
        """
        # Read the configuration file
        self.config = NexoclomConfig()
        # self._savepath = NexoclomConfig().savepath

        # Read in the input file:
        self._inputfile = infile
        params = []
        if os.path.isfile(infile):
            # Remove everything in the line after a comment character
            for line in open(infile, 'r'):
                if ';' in line:
                    line = line[:line.find(';')]
                elif '#' in line:
                    line = line[:line.find('#')]
                else:
                    pass
                    
                if line.count('=') == 1:
                    param_, val_ = line.split('=')
                    if param_.count('.') == 1:
                        sec_, par_ = param_.split('.')
                        params.append((sec_.casefold().strip(),
                                       par_.casefold().strip(),
                                       val_.strip()))
                    else:
                        pass
                else:
                    pass
        else:
            raise FileNotFoundError(infile)
            
        def extract_param(tag):
            return {b: c for (a, b, c) in params if a == tag}

        self.geometry = Geometry(extract_param('geometry'))
        self.surfaceinteraction = SurfaceInteraction(extract_param(
            'surfaceinteraction'))
        self.forces = Forces(extract_param('forces'))
        self.spatialdist = SpatialDist(extract_param('spatialdist'))
        self.speeddist = SpeedDist(extract_param('speeddist'))
        self.angulardist = AngularDist(extract_param('angulardist'))
        self.options = Options(extract_param('options'))
        
    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        else:
            return all([self.geometry == other.geometry,
                        self.surfaceinteraction == other.surfaceinteraction,
                        self.forces == other.forces,
                        self.spatialdist == other.spatialdist,
                        self.speeddist == other.speeddist,
                        self.angulardist == other.angulardist,
                        self.options == other.options])
        
    def __repr__(self):
        return self.__str__()

    def __str__(self):
        result = (self.geometry.__str__() + '\n' +
                  self.surfaceinteraction.__str__() + '\n' +
                  self.forces.__str__() + '\n' +
                  self.spatialdist.__str__() + '\n' +
                  self.speeddist.__str__() + '\n' +
                  self.angulardist.__str__() + '\n' +
                  self.options.__str__())
        
        return result

    def search(self):
        """ Search the database for previous model runs with the same inputs.
        See :doc:`searchtolerances` for tolerances used in searches.
        
        **Parameters**
        
        No parameters.
        
        **Returns**
        
        * A list of filenames corresponding to the inputs.
        
        * Number of packets contained in those saved outputs.
        
        * Total modeled source rate.
        """
        geo_id = self.geometry.search()
        sint_id = self.surfaceinteraction.search()
        for_id = self.forces.search()
        spat_id = self.spatialdist.search()
        spd_id = self.speeddist.search()
        ang_id = self.angulardist.search()
        opt_id = self.options.search()
        
        if None in [geo_id, sint_id, for_id, spat_id, spd_id, ang_id, opt_id]:
            return [], [], 0., 0.
        else:
            engine = self.config.create_engine()
            metadata_obj = sqla.MetaData()
            
            table = sqla.Table("outputfile", metadata_obj, autoload_with=engine)
            query = sqla.select(table).where(
                table.columns.geo_type == self.geometry.type,
                table.columns.geo_id == geo_id,
                table.columns.sint_type == self.surfaceinteraction.sticktype,
                table.columns.sint_id == sint_id,
                table.columns.force_id == for_id,
                table.columns.spatdist_type == self.spatialdist.type,
                table.columns.spatdist_id == spat_id,
                table.columns.spddist_type == self.speeddist.type,
                table.columns.spddist_id == spd_id,
                table.columns.angdist_type == self.angulardist.type,
                table.columns.angdist_id == ang_id,
                table.columns.opt_id == opt_id)
            
            with self.config.create_engine().connect() as con:
                result = pd.DataFrame(con.execute(query))
            
            if len(result) > 0:
                return (result.idnum.to_list(), result.filename.to_list(),
                        result.npackets.sum(), result.totalsource.sum())
            else:
                return [], [], 0, 0


    def run(self, npackets, packs_per_it=None, overwrite=False, compress=True):
        """Run the nexoclom model with the current inputs.
        
        **Parameters**
        
        npackets
            Number of packets to simulate
        
        packs_per_it
            Maximum number of packets to run at one time. Default = 1e5 in
            constant step-size mode; 1e6 in adaptive step-size mode.
        
        overwrite
            Erase any files matching the current inputs that exist.
            Default = False
            
        compress
            Remove packets with frac=0 from the outputs to reduce file size.
            Default = True
            
        **Outputs**
        
        Nothing is returned, but model runs are saved and cataloged.
        """
        t0_ = Time.now()
        print(f'Starting at {t0_}')

        # Determine how many packets have already been run
        if overwrite:
            self.delete_files()
            totalpackets = 0
        else:
            _, outputfiles, totalpackets, _ = self.search()
            print(f'Found {len(outputfiles)} files with {totalpackets} '
                  'packets.')

        npackets = int(npackets)
        ntodo = npackets - totalpackets
        
        if ntodo > 0:
            if (packs_per_it is None) and (self.options.step_size == 0):
                packs_per_it = 1000000
            elif packs_per_it is None:
                packs_per_it = (1e8 * self.options.step_size /
                                self.options.endtime.value)
            else:
                pass
            packs_per_it = int(np.min([ntodo, packs_per_it]))
            
            # Determine how many iterations are needed
            nits = int(np.ceil(ntodo/packs_per_it))
            
            print('Running Model')
            print(f'Will complete {nits} iterations of {packs_per_it} packets.')

            for _ in range(nits):
                tit0_ = Time.now()
                print(f'Starting iteration #{_+1} of {nits}')

                # Create an output object
                Output(self, packs_per_it, compress=compress)
                tit1_ = Time.now()
                print(f'Completed iteration #{_+1} in '
                      f'{(tit1_ - tit0_).sec} seconds.')
        else:
            pass

        t2_ = Time.now()
        dt_ = (t2_-t0_).sec
        if dt_ < 60:
            dt_ = f'{dt_} sec'
        elif dt_ < 3600:
            dt_ = f'{dt_/60} min'
        else:
            dt_ = f'{dt_/3600} hr'
        print(f'Model run completed in {dt_} at {t2_}.')

    def produce_image(self, format_, filenames=None, overwrite=False):
        return ModelImage(self, format_, filenames=filenames,
                          overwrite=overwrite)
    
    def delete_files(self):
        """Delete output files and remove them from the database.

        **Parameters**

        filelist
            List of files to remove. This can be found with Inputs.search()

        **Returns**

        No outputs.

        """
        idnum, filelist, _, _ = self.search()
        engine = self.config.create_engine()
        metadata_obj = sqla.MetaData()
        outputfile = sqla.Table("outputfile", metadata_obj, autoload_with=engine)
        modelimages = sqla.Table("modelimages", metadata_obj, autoload_with=engine)
        uvvsmodels = sqla.Table("uvvsmodels", metadata_obj, autoload_with=engine)
        spatdist = sqla.Table("spatdist_fittedoutput", metadata_obj,
                              autoload_with=engine)
        speeddist = sqla.Table("speeddist_fittedoutput", metadata_obj,
                              autoload_with=engine)
        
        for outid, outfile in zip(idnum, filelist):
            # Remove from database and delete file
            out_del = sqla.delete(outputfile).where(
                outputfile.columns.idnum == outid)
            
            # Delete any model images that depend on this output
            mod_select = sqla.select(modelimages).where(
                modelimages.columns.out_idnum == outid)
            mod_del = sqla.delete(modelimages).where(
                modelimages.columns.out_idnum == outid)

            # Delete any uvvs models that depend on this output
            uvvs_select = sqla.select(uvvsmodels).where(
                uvvsmodels.columns.out_idnum == outid)
            uvvs_del = sqla.delete(modelimages).where(
                uvvsmodels.columns.out_idnum == outid)
            
            # Delete any fitted uvvs models that depend on this output
            uvvsfit_select = sqla.select(uvvsmodels).where(
                uvvsmodels.columns.unfit_idnum == outid)
            uvvsfit_del = sqla.delete(modelimages).where(
                uvvsmodels.columns.unfit_idnum == outid)

            # Delete any fitted outputs that depend on this output
            fitted_spat_select = sqla.select(spatdist.columns.idnum).where(
                spatdist.columns.unfit_outid == outid)
            fitted_spat_delete = sqla.delete(spatdist).where(
                spatdist.columns.unfit_outid == outid)
            
            fitted_speed_select = sqla.select(speeddist.columns.idnum).where(
                speeddist.columns.unfit_outid == outid)
            fitted_speed_delete = sqla.delete(speeddist).where(
                speeddist.columns.unfit_outid == outid)
            
            fitted_out_spat_select = sqla.select(outputfile).where(
                outputfile.columns.spatdist_id.in_(fitted_spat_select),
                outputfile.columns.spatdist_type == 'fitted output')
            fitted_out_spat_select_ = sqla.select(outputfile.columns.idnum).where(
                outputfile.columns.spatdist_id.in_(fitted_spat_select),
                outputfile.columns.spatdist_type == 'fitted output')
            fitted_out_spat_delete = sqla.select(outputfile).where(
                outputfile.columns.spatdist_id.in_(fitted_spat_select),
                outputfile.columns.spatdist_type == 'fitted output')
            
            fitted_out_speed_select = sqla.select(outputfile).where(
                outputfile.columns.spddist_id.in_(fitted_speed_select),
                outputfile.columns.spddist_type == 'fitted output')
            fitted_out_speed_select_ = sqla.select(outputfile.columns.idnum).where(
                outputfile.columns.spddist_id.in_(fitted_speed_select),
                outputfile.columns.spddist_type == 'fitted output')
            fitted_out_speed_delete = sqla.select(outputfile).where(
                outputfile.columns.spddist_id.in_(fitted_speed_select),
                outputfile.columns.spddist_type == 'fitted output')
            
            # Delete any model images that depend on the fitted outputs
            fitted_images_spat_select = sqla.select(modelimages).where(
                modelimages.columns.out_idnum.in_(fitted_out_spat_select_))
            fitted_images_spat_delete = sqla.delete(modelimages).where(
                modelimages.columns.out_idnum.in_(fitted_out_spat_select_))

            fitted_images_speed_select = sqla.select(modelimages).where(
                modelimages.columns.out_idnum.in_(fitted_out_speed_select_))
            fitted_images_speed_delete = sqla.delete(modelimages).where(
                modelimages.columns.out_idnum.in_(fitted_out_speed_select_))

            with engine.connect() as con:
                mod_files = con.execute(mod_select)
                uvvs_files = con.execute(uvvs_select)
                uvvs_fit_files = con.execute(uvvsfit_select)
                fitted_out_spat_files = con.execute(fitted_out_spat_select)
                fitted_out_speed_files = con.execute(fitted_out_speed_select)
                fitted_images_spat_files = con.execute(fitted_images_spat_select)
                fitted_images_speed_files = con.execute(fitted_images_speed_select)
                
                con.execute(out_del)
                con.execute(mod_del)
                con.execute(uvvs_del)
                con.execute(uvvsfit_del)
                con.execute(fitted_spat_delete)
                con.execute(fitted_speed_delete)
                con.execute(fitted_out_spat_delete)
                con.execute(fitted_out_speed_delete)
                con.execute(fitted_images_spat_delete)
                con.execute(fitted_images_speed_delete)
                con.commit()
                
            if os.path.exists(outfile):
                print(f'Removing file {outfile}')
                os.remove(outfile)
                
            for row in mod_files:
                if os.path.exists(row.filename):
                    print(f'Removing file {row.filename}')
                    os.remove(row.filename)
                    
            for row in uvvs_files:
                if os.path.exists(row.filename):
                    print(f'Removing file {row.filename}')
                    os.remove(row.filename)
                    
            for row in uvvs_fit_files:
                if os.path.exists(row.filename):
                    print(f'Removing file {row.filename}')
                    os.remove(row.filename)

            for row in fitted_out_spat_files:
                if os.path.exists(row.filename):
                    print(f'Removing file {row.filename}')
                    os.remove(row.filename)

            for row in fitted_out_speed_files:
                if os.path.exists(row.filename):
                    print(f'Removing file {row.filename}')
                    os.remove(row.filename)

            for row in fitted_images_spat_files:
                if os.path.exists(row.filename):
                    print(f'Removing file {row.filename}')
                    os.remove(row.filename)
                    
            for row in fitted_images_speed_files:
                if os.path.exists(row.filename):
                    print(f'Removing file {row.filename}')
                    os.remove(row.filename)



# import sqlalchemy as sqla
# from nexoclom.utilities import NexoclomConfig
#
# def create_tables():
#     config = NexoclomConfig()
#     metadata_obj = sqla.MetaData()
#
#     geometry_with_time = sqla.Table(
#         '',
#         metadata_obj,
#         sqla.Column('idnum', sqla.Integer, sqla.Identity(), primary_key=True),
#
#         sqla.Column('name', sqla.String(30)),
#         sqla.Column('fullname', sqla.String)
# )
#
#
#
# # CREATE TYPE SSObject AS ENUM (
# #     'Milky Way', 'Sun', 'Mercury', 'Venus', 'Earth', 'Mars', 'Jupiter',
# #     'Saturn', 'Uranus', 'Neptune', 'Ceres', 'Pluto', 'Moon', 'Phobos',
# #     'Deimos', 'Io', 'Europa', 'Ganymede', 'Callisto', 'Mimas', 'Enceladus',
# #     'Tethys', 'Dione', 'Rhea', 'Titan', 'Hyperion', 'Iapetus', 'Phoebe',
# #     'Charon', 'Nix', 'Hydra'
# # );
#
# CREATE TABLE geometry_with_time (
#     idnum INT PRIMARY KEY GENERATED ALWAYS AS IDENTITY,
#     Planet SSObject NOT NULL,
#     StartPoint SSObject NOT NULL,
#     objects SSObject[],
#     starttime TIMESTAMP NOT NULL
# );
#
# CREATE TABLE geometry_without_time (
#     idnum INT PRIMARY KEY GENERATED ALWAYS AS IDENTITY,
#     Planet SSObject NOT NULL,
#     StartPoint SSObject NOT NULL,
#     objects SSObject ARRAY NOT NULL,
#     phi DOUBLE PRECISION ARRAY,
#     subsolarpt DOUBLE PRECISION[2] NOT NULL,
#     TAA DOUBLE PRECISION NOT NULL,
#     check (cardinality(objects)-1 = cardinality(phi)),
#     check (TAA >= 0 and TAA <= 6.284)
#     -- can add constraint on subsolarpoint and phi
# );
#
# CREATE TABLE surface_int_constant (
#     idnum INT PRIMARY KEY GENERATED ALWAYS AS IDENTITY,
#     stickcoef DOUBLE PRECISION NOT NULL,
#     accomfactor DOUBLE PRECISION,
#     check (stickcoef >= 0 and stickcoef <= 1),
#     check (accomfactor >= 0 and accomfactor <= 1)
# );
#
# CREATE TABLE surface_int_map (
#     idnum INT PRIMARY KEY GENERATED ALWAYS AS IDENTITY,
#     mapfile TEXT NOT NULL,
#     accomfactor DOUBLE PRECISION NOT NULL,
#     check (accomfactor >= 0 and accomfactor <= 1)
# );
#
# CREATE TABLE surface_int_tempdependent (
#     idnum INT PRIMARY KEY GENERATED ALWAYS AS IDENTITY,
#     accomfactor DOUBLE PRECISION NOT NULL,
#     a DOUBLE PRECISION[3] NOT NULL,
#     check (accomfactor >= 0 and accomfactor <= 1)
# );
#
# CREATE TABLE forces (
#     idnum INT PRIMARY KEY GENERATED ALWAYS AS IDENTITY,
#     gravity BOOLEAN NOT NULL,
#     radpres BOOLEAN NOT NULL
# );
#
# CREATE TABLE spatdist_uniform (
#     idnum INT PRIMARY KEY GENERATED ALWAYS AS IDENTITY,
#     exobase DOUBLE PRECISION NOT NULL,
#     longitude DOUBLE PRECISION[2] NOT NULL,
#     latitude DOUBLE PRECISION[2] NOT NULL,
#     check(exobase >= 1)
# );
#
# CREATE TABLE spatdist_surfmap (
#     idnum INT PRIMARY KEY GENERATED ALWAYS AS IDENTITY,
#     exobase DOUBLE PRECISION NOT NULL,
#     mapfile TEXT NOT NULL,
#     subsolarlon DOUBLE PRECISION,
#     coordinate_system TEXT NOT NULL,
#     check(exobase >= 1)
# );
#
# CREATE TABLE spatdist_spot (
#     idnum INT PRIMARY KEY GENERATED ALWAYS AS IDENTITY,
#     exobase DOUBLE PRECISION NOT NULL,
#     longitude DOUBLE PRECISION NOT NULL,
#     latitude DOUBLE PRECISION NOT NULL,
#     sigma DOUBLE PRECISION NOT NULL,
#     check(exobase >= 1)
# );
#
# CREATE TABLE spatdist_fittedoutput (
#     idnum INT PRIMARY KEY GENERATED ALWAYS AS IDENTITY,
#     unfit_outid INT NOT NULL,
#     query TEXT NOT NULL
# )
#
# CREATE TABLE speeddist_gaussian (
#     idnum INT PRIMARY KEY GENERATED ALWAYS AS IDENTITY,
#     vprob DOUBLE PRECISION NOT NULL,
#     sigma DOUBLE PRECISION NOT NULL,
#     check (sigma >= 0.)
# );
#
# CREATE TABLE speeddist_maxwellian (
#     idnum INT PRIMARY KEY GENERATED ALWAYS AS IDENTITY,
#     temperature DOUBLE PRECISION NOT NULL,
#     check (temperature >= 0)
# );
#
# CREATE TABLE speeddist_sputtering (
#     idnum INT PRIMARY KEY GENERATED ALWAYS AS IDENTITY,
#     alpha DOUBLE PRECISION NOT NULL,
#     beta DOUBLE PRECISION NOT NULL,
#     U DOUBLE PRECISION NOT NULL
# );
#
# CREATE TABLE speeddist_flat (
#     idnum INT PRIMARY KEY GENERATED ALWAYS AS IDENTITY,
#     vprob DOUBLE PRECISION NOT NULL,
#     delv DOUBLE PRECISION NOT NULL,
#     check (delv >= 0.)
# );
#
# CREATE TABLE speeddist_fittedoutput (
#     idnum INT PRIMARY KEY GENERATED ALWAYS AS IDENTITY,
#     unfit_outid INT NOT NULL,
#     query TEXT NOT NULL
# )
#
# CREATE TABLE speeddist_user (
#     idnum INT PRIMARY KEY GENERATED ALWAYS AS IDENTITY,
#     vdistfile TEXT NOT NULL)
#
# CREATE TABLE angdist_isotropic (
#     idnum INT PRIMARY KEY GENERATED ALWAYS AS IDENTITY,
#     altitude DOUBLE PRECISION[2] NOT NULL,
#     azimuth DOUBLE PRECISION[2] NOT NULL
# );
#
# CREATE TABLE options (
#     idnum INT PRIMARY KEY GENERATED ALWAYS AS IDENTITY,
#     endtime DOUBLE PRECISION NOT NULL,
#     species TEXT NOT NULL,
#     lifetime DOUBLE PRECISION NOT NULL,
#     outer_edge DOUBLE PRECISION NOT NULL,
#     step_size DOUBLE PRECISION NOT NULL,
#     resolution DOUBLE PRECISION,
#     fitted BOOLEAN NOT NULL,
#     check (endtime > 0),
#     check (outer_edge > 0),
#     check (step_size >= 0)
# );
#
# CREATE TABLE outputfile (
#     idnum INT PRIMARY KEY GENERATED ALWAYS AS IDENTITY,
#     filename TEXT UNIQUE NOT NULL,
#     npackets INT NOT NULL,
#     totalsource DOUBLE PRECISION NOT NULL,
#     creation_time TIMESTAMP NOT NULL,
#     geo_type TEXT NOT NULL,  --"with time" or "without time"
#     geo_id INT NOT NULL,
#     sint_type TEXT NOT NULL, --"constant", "temperature dependent", "from map"
#     sint_id INT NOT NULL,
#     force_id INT NOT NULL,
#     spatdist_type TEXT NOT NULL,
#     spatdist_id INT NOT NULL,
#     spddist_type TEXT NOT NULL,
#     spddist_id INT NOT NULL,
#     angdist_type TEXT NOT NULL,
#     angdist_id INT NOT NULL,
#     opt_id INT NOT NULL,
#     check (npackets > 0),
#     check (totalsource > 0)
# );
#
# CREATE TABLE modelimages (
#     idnum INT PRIMARY KEY GENERATED ALWAYS AS IDENTITY,
#     out_idnum INT NOT NULL,
#     quantity TEXT NOT NULL,
#     origin TEXT NOT NULL,
#     dims INT[2] NOT NULL,
#     center DOUBLE PRECISION[2] NOT NULL,
#     width DOUBLE PRECISION[2] NOT NULL,
#     subobslongitude DOUBLE PRECISION NOT NULL,
#     subobslatitude DOUBLE PRECISION NOT NULL,
#     mechanism TEXT,
#     wavelength TEXT,
#     filename TEXT NOT NULL)
#
# CREATE TABLE uvvsmodels (
#     idnum INT PRIMARY KEY GENERATED ALWAYS AS IDENTITY,
#     out_idnum INT NOT NULL,
#     unfit_idnum INT,
#     quantity TEXT NOT NULL,
#     query TEXT NOT NULL,
#     dphi DOUBLE PRECISION NOT NULL,
#     mechanism TEXT,
#     wavelength TEXT,
#     fitted BOOLEAN NOT NULL,
#     filename TEXT NOT NULL)
#
# CREATE TABLE savedpackets (
#     idnum INT PRIMARY KEY GENERATED ALWAYS AS IDENTITY,
#     query TEXT NOT NULL,
#     outputfile TEXT NOT NULL,
#     specind INT NOT NULL,
#     oint INT NOT NULL,
#     weight DOUBLE PRECISION NOT NULL,
#     frac0 DOUBLE PRECISION NOT NULL,
#     index0 INT NOT NULL,
#     ratio DOUBLE PRECISION NOT NULL,
#     scale_factor DOUBLE PRECISION)
#
# DONE



import numpy as np
from nexoclom.atomicdata import atomicmass
import astropy.units as u
import astropy.constants as const


def sputdist(velocity, U, alpha, beta, species):
    mspecies = atomicmass(species)
    v_b = np.sqrt(2*U/mspecies)
    v_b = v_b.to(u.km/u.s)
    f_v = velocity**(2*beta+1) / (velocity**2 + v_b**2)**alpha
    f_v /= np.max(f_v)
    return f_v.value


def MaxwellianDist(velocity, temperature, species):
    vth2 = 2*temperature*const.k_B/atomicmass(species)
    vth2 = vth2.to(u.km**2/u.s**2)
    f_v = velocity**3 * np.exp(-velocity**2/vth2)
    f_v /= np.max(f_v)
    return f_v.value

class CumDist:
    """Cumulative distribution function"""
    
    def __init__(self, x, y=None):
        """
        :param x: If y is specified, x is the values at which the probability
            distribution function was sampled;
            otherwise it is randomly chosen values from a distribution function
            
        :param y: If given, the probability distribution function
        """
        ind = np.argsort(x)
        self.x = x[ind]
        if y is None:
            self.sum = np.linspace(0, 1, len(x))
        else:
            y_ = y[ind]
            s = y_.cumsum()
            s -= s.min()
            self.sum = s/s.max()

# class CumDist2d:
#     """See Numerical Recipes, Ch 14.8"""
#     def __init__(self, x, y, sourcemap=None):
#         # x, y are randomly chosen from 2D distribution
#         q1, q2, q3, q4 = tuple(np.zeros((len(x), len(y))) for _ in range(4))
#         npts = len(x) * len(y)
#         grid = np.indices((len(x), len(y)))
#         xgrid, ygrid = grid[0], grid[1]
#
#         # q1 = (x[xgrid] <= x[np.newaxis,:]) & (y[ygrid] <= y[:,np.newaxis])
#
#         xLT = x[xgrid] <= x[np.newaxis,:]
#         yLT = y[ygrid] <= y[:,np.newaxis]
#         xGT = x[xgrid] > x[np.newaxis,:]
#         yGT = y[ygrid] > y[:,np.newaxis]
#
#         func_q1 = lambda i, j: sum(xLT[:,i] & yLT[j,:])
#         from IPython import embed; embed()
#         import sys; sys.exit()
#
#         q1 = np.fromfunction(func_q1, (10, 10), dtype=int)
#
#         # print(q1pts[5, 5])
#         i, j = 5, 5
#         print(q1[i,j])
#         print(np.sum((x <= x[i]) & (y <= y[j])))
#
#
#
#         if sourcemap is None:
#             for i in range(len(x)):
#                 for j in range(len(y)):
#                     q1[i,j] = np.sum((x <= x[i]) & (y <= y[j]))/npts
#                     q2[i,j] = np.sum((x <= x[i]) & (y > y[j]))/npts
#                     q3[i,j] = np.sum((x > x[i]) & (y <= y[j]))/npts
#                     q4[i,j] = np.sum((x > x[i]) & (y > y[j]))/npts
#         else:
#             # func = interpolate.RectBivariateSpline(sourcemap['longitude'].value,
#             #                                        sourcemap['latitude'].value,
#             #                                        sourcemap['map'])
#             # z = func.ev(x, y)
#             x, y = sourcemap['longitude'].value, sourcemap['latitude'].value
#             z = sourcemap['map']
#             total = np.sum(z)
#             for i in range(len(x)):
#                 for j in range(len(y)):
#                     q1[i] = np.sum(z[(x <= x[i]) & (y <= y[i])])/total
#                     q2[i] = np.sum(z[(x <= x[i]) & (y > y[i])])/total
#                     q3[i] = np.sum(z[(x > x[i]) & (y <= y[i])])/total
#                     q4[i] = np.sum(z[(x > x[i]) & (y > y[i])])/total



"""Read atomicdata from the text files and save as pandas dataframes
"""
import os
import glob
from nexoclom import __file__ as basefile
import pandas as pd
from nexoclom.atomicdata import gValue, PhotoRate

basepath = os.path.dirname(basefile)

def make_gvalue_table():
    ref = 'Killen et al. (2009)'
    datafiles = glob.glob(os.path.join(basepath, 'data', 'g-values', '*.dat'))
    
    gvalues = pd.DataFrame(columns=['species', 'wavelength', 
                                    'velocity', 'gvalue', 'refpoint', 
                                    'filename', 'reference'])
    
    for datafile in datafiles:
        # Determine the species
        species = os.path.basename(datafile).split('.')[0]

        with open(datafile) as f:
            # Determine the reference point
            refpt_str = f.readline().strip()
        refpt = float(refpt_str.split('=')[1])

        gvalue_species = pd.read_csv(datafile, sep=':', skiprows=1)
        wavelengths = [float(wave) for wave in gvalue_species.columns[1:]]
        gvalue_species.columns = ['vel'] + wavelengths

        for wave in wavelengths:
            print(species, wave)
            for _, row in gvalue_species.iterrows():
                newrow = {'species': species,
                          'wavelength': wave,
                          'velocity': row['vel'],
                          'gvalue': row[wave],
                          'refpoint': refpt,
                          'filename': datafile,
                          'reference': ref}
                gvalues.loc[len(gvalues)] = newrow

    fakerow = {'species': 'Fk',
                'wavelength': 9999,
                'velocity': 0., 
                'gvalue': 0.,
                'refpoint': 1.,
                'filename': 'NO_SUCH_FILE.dat',
                'reference': 'Fake (2021)'}
    gvalues.loc[len(gvalues)] = fakerow

    fakerow = {'species': 'Fk',
                'wavelength': 9999,
                'velocity': 0., 
                'gvalue': 1., 
                'refpoint': 1.,
                'filename': 'NO_SUCH_FILE2.dat',
                'reference': 'Liar (2021)'}
    gvalues.loc[len(gvalues)] = fakerow
    
    gvalue_file = gValue.gvalue_filename()
    print(gvalue_file)
    gvalues.to_pickle(gvalue_file)

def make_photorates_table():
    photodatafiles = glob.glob(os.path.join(basepath, 'data', 'Loss', 
                                            'Photo', '*.dat'))
    species, reaction, kappa, reference, best = [], [], [], [], []
    for photofile in photodatafiles:
        print(f'  {photofile}')
        ref = ''
        for line in open(photofile):
            if 'reference' in line.lower():
                ref = line.split('//')[0].strip()
            elif len(line.split(':')) == 4:
                parts = line.split(':')
                species.append(parts[0].strip())
                reaction.append(parts[1].strip())
                kappa.append(float(parts[2].strip()))
                reference.append(ref)
                best.append(True)
            else:
                pass
    photorates = pd.DataFrame({'species': species,
                               'reaction': reaction,
                               'kappa': kappa,
                               'reference': reference,
                               'best_version': best}) 
    
    # Find duplicates
    counts = photorates['reaction'].value_counts()
    duplicates = counts[counts > 1].index.values
    for duplicate in duplicates:
        subset = photorates[photorates.reaction == duplicate]
        print(f'Reaction = {duplicate}')
        for ind, row in subset.iterrows():
            print(f'{ind}: {row.reference}')
        which = input('Which is the best source? ')
        notwhich = [i for i in subset.index if i != which]
        photorates.loc[notwhich, 'best_version'] = False

    photorates.to_pickle(PhotoRate.photorates_filename())


from nexoclom.math import minmaxmean
import numpy as np
import pytest


@pytest.mark.math
def test_minmaxmean():
    array = np.random.random(50)
    assert (np.nanmin(array), np.nanmax(array), np.nanmean(array)) == minmaxmean(array)


import numpy as np
from ..atomicdata import PhotoRate
import astropy.units as u

class LossInfo:
    def __init__(self, atom, lifetime, aplanet):
        # Initialization
        self.photo = 0.
        self.eimp = 0.
        self.chX = 0.
        self.reactions = []
        
        if isinstance(lifetime, type(1*u.s)):
            lifetime_ = lifetime.value
        else:
            lifetime_ = lifetime

        if lifetime_ < 0:
            self.photo = np.abs(1./lifetime_)
            self.reactions = 'Generic photo reaction'
        elif lifetime_ == 0:
            photo = PhotoRate(atom, aplanet)
            self.photo = photo.rate.value
            self.reactions = photo.reactions['reaction'].values

            # Electron impact

            # Charge exchange
        else:
            print('LossInfo objects should not be '
                  'instantiated with lifetime > 0')
            pass

        if len(self.reactions) == 0:
            self.reactions = None

    def __len__(self):
        return len(self.reactions) if self.reactions is not None else 0

    def __str__(self):
        if len(self) == 0:
            return 'No reactions included'
        elif len(self) == 1:
            result = f'Included Reaction: {self.reactions[0]}'
        else:
            reacs = '\n\t'.join(self.reactions)
            result = f'Included Reactions: {reacs}'

        if self.photo != 0:
            result += f'\nPhoto Rate = {self.photo:0.2e} s'
        if self.eimp != 0:
            result += f'\nElectron Impact Rate = {self.eimp:0.2e} UNIT'
        if self.chX!= 0:
            result += f'\nCharge Exchange Rate = {self.chX:0.2e} UNIT'

        return result



import os.path
import numpy as np
import copy
import astropy.units as u
from sklearn.neighbors import KDTree
from nexoclom import math as mathMB
from nexoclom.atomicdata import gValue
from nexoclom.modelcode.input_classes import InputError
from nexoclom.modelcode.Output import Output


class ModelResult:
    """Base class for nexoclom model comparisons with data.

    The ModelResult object is the base class for ModelImage (radiance and column
    density images), LOSResult (radiance or column along lines of sight), and
    ModelDensity (density along a trajectory or plane - not written yet).
    
    **Parameters**
    
    inputs
        An Input object
        
    params
        A dictionary containing parameters needed to create the model result.
        See LOSResult.py or ModelImage.py for details.
    
    **Methods**
    
    packet_weighting
        Determine the weighting factor each packet. When determining density
        or column, no weighting is needed. For radiance, takes into account
        determines g-value and takes into account shadowing.
        
    transform_reference_frame
        Will be used to transform packets from the central object reference frame
        to another location
        
    **Attributes**
    
    inputs
        Input object with the parameters associated with the model result
        
    params
        A dictionary containing parameters of the result. See LOSResult.py
        or ModelImage.py for details.
        
    outputfiles
        locations of saved Outputs associated with the inputs
    
    npackets
        total number of packets simulated
    
    totalsource
        total source in packets (if the initial fraction are all 1, then
        totalsource = npackets * nsteps
        
    quantity
        column, density, or radiance determined from params
        
    mechanism
        Emission mechanism if quantity = radiance else None
        
    wavelength
        Emssion wavelengths if quantity = radiance else None
      
    """
    def __init__(self, inputs, params):
        """
        :param inputs: Input object
        :param params: Dictionary with ModelResult parameters
        """
        self.inputs = copy.deepcopy(inputs)
        self.outid, self.outputfiles, _, _ = self.inputs.search()
        self.npackets = 0
        self.totalsource = 0.
        self.atoms_per_packet = 0.
        self.sourcerate = 0.
        if isinstance(params, str):
            if os.path.exists(params):
                self.params = {}
                with open(params, 'r') as f:
                    for line in f:
                        if ';' in line:
                            line = line[:line.find(';')]
                        elif '#' in line:
                            line = line[:line.find('#')]
                        else:
                            pass
                        
                        if '=' in line:
                            p, v = line.split('=')
                            self.params[p.strip().lower()] = v.strip()
                        else:
                            pass
            else:
                raise FileNotFoundError('ModelResult.__init__',
                                        'params file not found.')
        elif isinstance(params, dict):
            self.params = params
        else:
            raise TypeError('ModelResult.__init__',
                            'params must be a dict or filename.')
            
        # Do some validation
        quantities = ['column', 'radiance', 'density']
        self.quantity = self.params.get('quantity', None)
        if (self.quantity is None) or (self.quantity not in quantities):
            raise InputError('ModelImage.__init__',
                             "quantity must be 'column' or 'radiance'")
        else:
            pass

        if self.quantity == 'radiance':
            # Note - only resonant scattering currently possible
            self.mechanism = ['resonant scattering']
    
            if 'wavelength' in self.params:
                self.wavelength = tuple(sorted(int(m.strip())*u.AA for m
                    in self.params['wavelength'].split(',')))
            elif self.inputs.options.species is None:
                raise InputError('ModelImage.__init__',
                                 'Must provide either species or params.wavelength')
            elif self.inputs.options.species == 'Na':
                self.wavelength = (5891*u.AA, 5897*u.AA)
            elif self.inputs.options.species == 'Ca':
                self.wavelength = (4227*u.AA,)
            elif self.inputs.options.species == 'Mg':
                self.wavelength = (2852*u.AA,)
            else:
                raise InputError('ModelResult.__init__', ('Default wavelengths '
                                 f'not available for {self.inputs.options.species}'))
        else:
            self.mechanism = None
            self.wavelength = None
            
        self.unit = u.def_unit('R_' + self.inputs.geometry.planet.object,
                               self.inputs.geometry.planet.radius)

    def packet_weighting(self, packets, out_of_shadow, aplanet):
        """
        Determine weighting factor for each packet
        :param packets: DataFrame with packet parameters
        :param out_of_shadow: Boolean array, True if in sunlight; False if in shadow
        :param aplanet: Distance of planet from Sun (used for g-value calculation)
        :return: Adds a 'weight' column to the packets DataFrame
        """
        if self.quantity == 'column':
            packets['weight'] = packets['frac']
        elif self.quantity == 'density':
            packets['weight'] = packets['frac']
        elif self.quantity == 'radiance':
            if 'resonant scattering' in self.mechanism:
                gg = np.zeros(len(packets))/u.s
                for w in self.wavelength:
                    gval = gValue(self.inputs.options.species, w, aplanet)
                    gg += mathMB.interpu(packets['radvel_sun'].values *
                                         self.unit/u.s, gval.velocity, gval.g)

                weight_resscat = packets['frac']*out_of_shadow*gg.value/1e6
            else:
                weight_resscat = np.zeros(len(packets))
                
            packets['weight'] = weight_resscat  # + other stuff
        else:
            raise InputError('ModelResults.packet_weighting',
                             f'{self.quantity} is invalid.')

        assert np.all(np.isfinite(packets['weight'])), 'Non-finite weights'

    def make_source_map(self, normalize=True):
        """
        At each point in lon/lat grid want:
            * Source flux (atoms/cm2/s
            * Speed distribution (f_v vs v)
            * Azimuthal distribution (f_az vs az) -> measured CCW from north
            * Altitude distribution (f_alt vs alt) -> tangent = 0, normal = 90
        """
        X0 = None
        for outputfile in self.outputfiles:
            output = Output.restore(outputfile)
            if X0 is None:
                X0 = output.X0[['x', 'y', 'z', 'vx', 'vy', 'vz', 'frac']]
            else:
                X0 = X0.append(output.X0[['x', 'y', 'z', 'vx', 'vy', 'vz', 'frac']],
                               ignore_index=True)
            del output

        velocity = (np.array([X0.vx.values, X0.vy.values, X0.vz.values]) *
                    self.inputs.geometry.planet.radius.value)
        speed = np.linalg.norm(velocity, axis=0)

        # Radial, east, north unit vectors
        rad = np.array([X0.x.values, X0.y.values, X0.z.values])
        rad_ = np.linalg.norm(rad, axis=0)
        rad = rad/rad_[np.newaxis, :]

        east = np.array([X0.y.values, -X0.x.values, np.zeros_like(X0.z.values)])
        n = np.zeros_like(rad)
        n[2,:] = 1
        east_ = np.linalg.norm(east, axis=0)
        east = east/east_[np.newaxis, :]

        # north = np.array([-X0.z.values * X0.x.values,
        #                   -X0.z.values * X0.y.values,
        #                   X0.x.values**2 + X0.y.values**2])
        north = np.cross(rad, east, axis=0)
        north_ = np.linalg.norm(north, axis=0)
        north = north/north_[np.newaxis, :]

        v_rad = np.sum(velocity * rad, axis=0)
        v_east = np.sum(velocity * east, axis=0)
        v_north = np.sum(velocity * north, axis=0)

        v_rad_over_speed = v_rad/speed
        v_rad_over_speed[v_rad_over_speed > 1] = 1
        v_rad_over_speed[v_rad_over_speed < -1] = -1

        assert np.all(np.isclose(v_rad**2 + v_east**2 + v_north**2, speed**2))
        X0['altitude'] = np.arcsin(v_rad_over_speed)
        X0['azimuth'] = (np.arctan2(v_north, v_east) + 2*np.pi) % (2*np.pi)
        X0['v_rad'] = v_rad
        X0['v_east'] = v_east
        X0['v_north'] = v_north
        X0['speed'] = speed
        X0['longitude'] = (np.arctan2(X0.x.values, -X0.y.values) + 2*np.pi) % (2*np.pi)
        X0['latitude'] = np.arcsin(X0.z.values)

        source = self._calculate_histograms(X0, normalize, weight=True)
        if  self.__dict__.get('fitted', False):
            available = self._calculate_histograms(X0, normalize, weight=False)
        else:
            available = None

        return source, available, X0
    
    def velocity_distribution_at_point(self, point, radius=5*np.pi/180, X0=None,
                                       nvelbins=100, nazbins=180, naltbins=45,
                                       normalize=True):
        if X0 is None:
            _, _, X0 = self.make_source_map()
        else:
            pass
        
        tree = KDTree(X0[['x', 'y', 'z']].values)
        point_xyz = np.array([np.sin(point[0]) * np.cos(point[1]),
                             -np.cos(point[0]) * np.cos(point[1]),
                             np.sin(point[1])]).reshape((1, 3))
        ind = tree.query_radius(point_xyz, radius)

        source_point = self._calculate_histograms(X0.iloc[ind[0]], normalize,
            weight=True, nlonbins=0, nlatbins=0, nvelbins=nvelbins, nazbins=nazbins,
            naltbins=naltbins)
        
        if self.__dict__.get('fitted', False):
            available_point = self._calculate_histograms(X0.iloc[ind[0]], normalize,
                weight=False, nlonbins=0, nlatbins=0, nvelbins=nvelbins,
                nazbins=nazbins, naltbins=naltbins)
        else:
            available_point = None

        return source_point, available_point

    def _calculate_histograms(self, X0, normalize, weight=False, nlonbins=72, nlatbins=36,
                             nvelbins=100, nazbins=180, naltbins=45):
        if weight:
            w = X0.frac.values
        else:
            w = np.ones_like(X0.frac.values)
    
        # Determine source distribution
        if (nlonbins > 0) and (nlatbins > 0):
            source = mathMB.Histogram2d(X0.longitude, X0.latitude, weights=w,
                                        range=[[0, 2*np.pi], [-np.pi/2, np.pi/2]],
                                        bins=(nlonbins, nlatbins))
            source.x, source.dx = source.x * u.rad, source.dx * u.rad
            source.y, source.dy = source.y * u.rad, source.dy * u.rad
        
            if normalize:
                # Convert histogram to flux
                # (a) divide by area of a grid cell
                #   Surface area of a grid cell = R**2 (lambda_2 - lambda_1) (sin(phi2)-sin(phi1))
                #   https://onlinelibrary.wiley.com/doi/epdf/10.1111/tgis.12636, eqn 1
                # (b) Multiply by source rate
                _, gridlatitude = np.meshgrid(source.x, source.y)
                area = (self.inputs.geometry.planet.radius**2 * source.dx *
                        (np.sin(gridlatitude + source.dy / 2) - np.sin(gridlatitude - source.dy / 2)))
            
                source.histogram = (source.histogram / X0.frac.sum() /
                                    area.T.to(u.cm**2) *
                                    self.sourcerate.to(1 / u.s))
            else:
                pass
        else:
            source = None
    
        # Velocity flux atoms/s/(km/s)
        if nvelbins > 0:
            velocity = mathMB.Histogram(X0.speed, bins=nvelbins,
                                        range=[0, X0.speed.max()], weights=w)
            velocity.x = velocity.x * u.km / u.s
            velocity.dx = velocity.dx * u.km / u.s
            if normalize:
                velocity.histogram = (self.sourcerate * velocity.histogram /
                                    velocity.histogram.sum() / velocity.dx)
                velocity.histogram = velocity.histogram.to(self.sourcerate.unit *
                                                           u.def_unit('(km/s)^-1', u.s/u.km))
            else:
                pass
        else:
            velocity = None
    
        # Altitude distribution
        if naltbins > 0:
            altitude = mathMB.Histogram(X0.altitude, bins=naltbins,
                                        range=[0, np.pi / 2], weights=w)
            altitude.x = altitude.x * u.rad
            altitude.dx = altitude.dx * u.rad
            if normalize:
                altitude.histogram = (self.sourcerate * altitude.histogram /
                                    altitude.histogram.sum() / altitude.dx)
            else:
                pass
        else:
            altitude = None
    
        # Azimuth distribution
        if nazbins > 0:
            azimuth = mathMB.Histogram(X0.azimuth, bins=nazbins,
                                       range=[0, 2 * np.pi], weights=w)
            azimuth.x = azimuth.x * u.rad
            azimuth.dx = azimuth.dx * u.rad
            if normalize:
                azimuth.histogram = (self.sourcerate * azimuth.histogram /
                                    azimuth.histogram.sum() / azimuth.dx)
            else:
                pass
        else:
            azimuth = None
    
        source = {'abundance':source,
                  'speed':velocity,
                  'altitude':altitude,
                  'azimuth':azimuth,
                  'coordinate_system':'solar-fixed'}
    
        return source
    
    def show_source_map(self, filename, which='source', smooth=False, show=True,
                        source=None, available=None, X0=None):
        if X0 is None:
            source, available, X0 = self.make_source_map()
        elif (which == 'source') and (source is None):
            touse, _, X0 = self.make_source_map()
        elif which == 'source':
            touse = source
        elif (which == 'available') and (available is None):
            _, touse, X0 = self.make_source_map()
        elif which == 'available':
            touse = available
        else:
            raise InputError
        
        
        

    # def transform_reference_frame(self, output):
    #     """If the image center is not the planet, transform to a
    #        moon-centric reference frame."""
    #     assert 0, 'Not ready yet.'
    #
    #     # Load output
    #
    #     # # Transform to moon-centric frame if necessary
    #     # if result.origin != result.inputs.geometry.planet:
    #     #     assert 0, 'Need to do transparamsion for a moon.'
    #     # else:
    #     #     origin = np.array([0., 0., 0.])*output.x.unit
    #     #     sc = 1.
    #
    #     # Choose which packets to use
    #     # touse = output.frac >= 0 if keepall else output.frac > 0
    #
    #     # packet positions relative to origin -- not rotated
    #     # pts_sun = np.array((output.x[touse]-origin[0],
    #     #                     output.y[touse]-origin[1],
    #     #                     output.z[touse]-origin[2]))*output.x.unit
    #     #
    #     # # Velocities relative to sun
    #     # vels_sun = np.array((output.vx[touse],
    #     #                      output.vy[touse],
    #     #                      output.vz[touse]))*output.vx.unit
    #
    #     # Fractional content
    #     # frac = output.frac[touse]
    #
    #     return output #, pts_sun, vels_sun, frac



from nexoclom.math.interpu import interpu
from nexoclom.math.minmaxmean import minmaxmean
from nexoclom.math.randomdeviates import random_deviates_1d, random_deviates_2d
from nexoclom.math.rotation_matrix import rotation_matrix
from nexoclom.math.smooth import smooth
from nexoclom.math.histogram import Histogram, Histogram2d
from nexoclom.math.distributions import sputdist, MaxwellianDist, CumDist


__name__ = 'math'
__author__ = 'Matthew Burger'
__email__ = 'mburger@stsci.edu'



from nexoclom.atomicdata.g_values import RadPresConst, gValue
from nexoclom.atomicdata.photolossrates import PhotoRate
from nexoclom.atomicdata.atomicmass import atomicmass


__name__ = 'atomicdata'
__author__ = 'Matthew Burger'
__email__ = 'mburger@stsci.edu'


import numpy as np


def surface_temperature(geometry, longitude, latitude,
                        t0=100., t1=None, n=.25):
    if geometry.startpoint == 'Mercury':
        if t1 is None:
            t1 = 600. + 125*(np.cos(geometry.taa) - 1)/2.
            t1 = t1.value
        else:
            pass
        
        t_surf = np.zeros_like(longitude) + t0
        mask = (longitude <= np.pi/2) | (longitude >= 3*np.pi/2)
        t_surf[mask] = t0 + t1*np.abs(np.cos(longitude[mask]) *
                                      np.cos(latitude[mask]))**n
        
        return t_surf


import os
import spiceypy as spice
from bs4 import BeautifulSoup
import requests
from nexoclom import __datapath__


def load_kernels():
    """Load generic spice kernels. Retrieves if necessary."""
    # leap second kernel
    url = 'https://naif.jpl.nasa.gov/pub/naif/generic_kernels/lsk/'
    page = requests.get(url).text
    soup = BeautifulSoup(page, 'html.parser')
    
    for node in soup.find_all('a'):
        text = node.get('href')
        if (text is not None) and (text.endswith('.tls')):
        
    lsk_url = [url + '/' + node.get('href')
               for node in soup.find_all('a')
               if (node.get('href') is not None) and
                  node.get('href').endswith('.tls')]
    lsk_path = os.path.join(__datapath__, 'spice_kernels', 'lsk')
    lsk_filename = os.path.join(lsk_path, os.path.basename(lsk_url[-1]))
    if not os.path.exists(lsk_filename):
        print(f'Retreiving leapsecond kernel {os.path.basename(lsk_filename)}')
        file = requests.get(lsk_url[-1]).text
        with open(lsk_filename, 'w') as f:
            f.write(file)
    else:
        pass
    
    spice.furnsh(lsk_filename)
    

    


"""``minmaxmean()``: Print np.min(), np.max(), np.mean() at once."""
import numpy as np


def minmaxmean(x):
    """Print np.min(), np.max(), np.mean() at once."""
    return np.nanmin(x), np.nanmax(x), np.nanmean(x)



"""Determine distance and radial velocity relative to Sun."""
import numpy as np
from scipy.misc import derivative
import astropy.units as u
# from astroquery.jplhorizons import Horizons
from nexoclom.solarsystem.SSObject import SSObject


def planet_dist(planet_, taa=None, time=None):
    if isinstance(planet_, str):
        planet = SSObject(planet_)
        
        if planet.object is None:
            return None
    elif isinstance(planet_, SSObject):
        planet = planet_
    else:
        raise TypeError('solarsystemMB.planet_dist',
                        'Must give a SSObject or a object name.')

    if time is not None:
        raise NotImplementedError
        # et = spice.str2et(time.isot)
        # posvel, lt = spice.spkezr(planet.object, et, 'J2000',
        #                           'LT+S', 'Sun')
        #
        # position = np.array(posvel[0:3])*u.km
        # r = np.sqrt(np.sum(position**2))
        #
        # velocity = np.array(posvel[3:])*u.km/u.s
        # v_r = np.sum(position*velocity)/r
        # r = r.to(u.au)
    elif taa is not None:
        a = planet.a
        eps = planet.e

        # make sure taa is in radians. If not a quantity, assume it is.
        if isinstance(taa, type(1*u.s)):
            taa_ = taa.to(u.rad).value
        elif type(taa) in (int, float):
            taa_ = taa
        else:
            raise TypeError('taa must be a number or angle quantity')

        if eps > 0:
            # determine r
            r = a * (1-eps**2)/(1+eps*np.cos(taa_))
            P = np.sqrt(a**3/(1*u.au)**3) * u.yr

            # determine v_r = dr/dt
            def trueanom(time):
                # Mean anomally
                M = 2*np.pi*time/P.value

                # Determine eccentric anomaly from M = E-e*sin(E)
                EEtemp = np.linspace(0, 2*np.pi, 1001)
                mm = EEtemp - eps*np.sin(EEtemp)
                EE = np.array([np.interp(x, mm, EEtemp) for x in M])

                # True anomaly
                phi = (2*np.arctan(np.sqrt((1+eps)/(1-eps)) * np.tan(EE/2)) +
                       (2*np.pi)) % (2*np.pi)
                return phi

            def radius(time):
                phi = trueanom(time)
                r = a * (1-eps**2)/(1+eps*np.cos(phi))
                return r.value

            time = np.linspace(-1, 1, 2001)*P.value
            radvel = derivative(radius, time, dx=1e-3)
            radvel = radvel.astype(np.float64)
            ttt = trueanom(time)
            v_r = np.interp(taa_, ttt, radvel, period=2*np.pi)
            v_r *= a.unit/P.unit
            v_r = v_r.to(u.km/u.s)
        else:
            r, v_r = a, 0.*u.km/u.s
    else:
        print('Neither a time nor a true anomaly was given.')
        return None

    return r, v_r



import os
import psycopg2
from sqlalchemy import create_engine
import subprocess
from nexoclom.utilities.exceptions import ConfigfileError


DEFAULT_DATABASE = 'thesolarsystemmb'
DEFAULT_PORT = 5432


class NexoclomConfig:
    """Configure external resources used in the model.
    The following parameters can be saved in the file `$HOME/.nexoclom`.
    * savepath = <path where output files are saved>
    * database = <name of the postgresql database to use> (*optional*)
    * port = <port for postgreSQL server to use> (*optional*)
    * dbhost = <hostname for postgreSQL database> (*optional* - leave blank
        to use local database)
    
    If savepath is not present, an exception is raised
    """
    def __init__(self, verbose=False):
        configfile = os.environ.get('NEXOCLOMCONFIG', os.path.join(
            os.environ['HOME'], '.nexoclom'))
        self.configfile = os.path.expandvars(configfile)
        
        if verbose:
            print(f'Using configuration file {self.configfile}')
        else:
            pass
        
        config = {}
        if os.path.isfile(os.path.expandvars(self.configfile)):
            # Read the config file into a dict
            for line in open(self.configfile, 'r'):
                if '=' in line:
                    key, value = line.split('=')
                    config[key.strip()] = value.strip()
                else:
                    pass
        else:
            pass
        
        self.savepath = config.get('savepath', None)
        if self.savepath is None:
            raise ConfigfileError(self.configfile, self.savepath)
        elif not os.path.exists(self.savepath):
            os.makedirs(self.savepath)
        else:
            pass
        
        self.database = config.get('database', DEFAULT_DATABASE)
        
        if 'port' not in config:
            self.port = DEFAULT_PORT
        else:
            self.port = int(config['port'])
            
        self.mesdatabase = config.get('mesdatabase', None)
        self.mesdatapath = config.get('mesdatapath', None)
        
        for key, value in config.items():
            if key not in self.__dict__:
                self.__dict__[key] = value
            else:
                pass
            
        self.dbhost = config.get('dbhost', None)
        
    def __repr__(self):
        return self.__dict__.__repr__()
    
    def __str__(self):
        return self.__dict__.__str__()
    
    def verify_database_running(self):
        # verify database is running; start it if it isn't
        proc = subprocess.run('pg_ctl status', capture_output=True, shell=True)
        if 'server is running' in str(proc.stdout):
            return 'Database Already Running'
        else:
            pg_log_dir = os.path.join(os.path.expandvars(os.environ['PGDATA']),
                                      'logfile')
            subprocess.run(f'pg_ctl -o "-p {self.port}" start -l {pg_log_dir}',
                           shell=True)
            return 'Started Database'

    def create_engine(self, database=None):
        """Wrapper for slalchemy.create_engine() that determines which database and port to use.
        :param database: Default = None to use value from config file
        :return: SQLAlchemy engine
        """
        self.verify_database_running()
        if database is None:
            database = self.database
        else:
            pass
        
        if self.dbhost:
            url = (f"postgresql+psycopg2://{os.environ['USER']}@{self.dbhost}:"
                   f"{self.port}/{database}")
        else:
            url = (f"postgresql+psycopg2://{os.environ['USER']}@localhost:"
                   f"{self.port}/{database}")
        engine = create_engine(url, echo=False, future=True)

        return engine

    def database_connect(self, database=None):
        self.verify_database_running()
        if database is None:
            database = self.database
        else:
            pass
        
        if self.dbhost:
            con = psycopg2.connect(host=self.dbhost, dbname=database,
                                   port=self.port)
        else:
            con = psycopg2.connect(dbname=database, port=self.port)
        con.autocommit = True

        return con



import os.path
import numpy as np
import pickle
import astropy.units as u
import json
import sqlalchemy as sqla
import sqlalchemy.dialects.postgresql as pg

from nexoclom.math import rotation_matrix, Histogram2d
from nexoclom.modelcode.ModelResult import ModelResult
from nexoclom.modelcode.Output import Output

import bokeh.plotting as bkp
from bokeh.palettes import Inferno256
from bokeh.models import (HoverTool, ColumnDataSource, ColorBar,
                          LogColorMapper, LogTicker, LinearColorMapper)
from bokeh.io import curdoc, export_png
from bokeh.themes import Theme


class ModelImage(ModelResult):
    def __init__(self, inputs, params, filenames=None, overwrite=False):
        """ Create Images from model results.
        This Assumes the model has already been run.
        
        Parameters
        ==========
        inputs
            An Input object
        
        params
            A dictionary with format information or a path to a formatfile.
            
        filenames
            A filename or list of filenames to use. Default = None is to
            find all files created for the inputs.
            
        overwrite
            If True, deletes any images that have already been computed.
            Default = False
        """
        super().__init__(inputs, params)
        self.type = 'image'
        self.origin = self.params.get('origin', inputs.geometry.planet)
        self.unit = u.def_unit('R_' + self.origin.object,
                               self.origin.radius)

        dimtemp = self.params.get('dims', '800,800').split(',')
        self.dims = [int(dimtemp[0]), int(dimtemp[1])]

        centtemp = self.params.get('center', '0,0').split(',')
        self.center = [float(centtemp[0])*self.unit,
                       float(centtemp[1])*self.unit]

        widtemp = self.params.get('width', '8,8').split(',')
        self.width = [float(widtemp[0])*self.unit,
                      float(widtemp[1])*self.unit]

        subobslong = self.params.get('subobslongitude', '0')
        self.subobslongitude = float(subobslong) * u.rad

        subobsllat = self.params.get('subobslatitude', np.pi/2)
        self.subobslatitude = float(subobsllat) * u.rad

        self.image = np.zeros(self.dims)
        self.packet_image = np.zeros(self.dims)
        self.blimits = None
        immin = tuple(c - w/2 for c, w in zip(self.center, self.width))
        immax = tuple(c + w/2 for c, w in zip(self.center, self.width))
        self.xrange = [immin[0], immax[0]]
        self.zrange = [immin[1], immax[1]]
        scale = tuple(w/d for w, d in zip(self.width, self.dims))
        self.Apix = (scale[0]*scale[1]).to(u.cm**2)

        self.xaxis = None
        self.zaxis = None

        self.outid, self.outputfiles, _, _ = self.inputs.search()
        
        for i, fname in enumerate(self.outputfiles):
            # Search to see if its already been done
            print(f'Output filename: {fname}')
            image_, packets_ = self.restore(fname, overwrite=overwrite)
            output = Output.restore(fname)

            if image_ is None:
                image_, packets_, = self.create_image(output)
                print(f'Completed image {i+1} of {len(self.outputfiles)}')
            else:
                print(f'Image {i+1} of {len(self.outputfiles)} '
                       'previously completed.')

            self.image += image_.histogram
            self.packet_image += packets_.histogram
            self.totalsource += output.totalsource
            self.xaxis = image_.x * self.unit
            self.zaxis = image_.y * self.unit
            del output

        mod_rate = self.totalsource / self.inputs.options.endtime.value
        self.atoms_per_packet = 1e23 / mod_rate
        self.sourcerate = 1e23 / u.s
        self.image *= self.atoms_per_packet

    def save(self, fname, image, packets):
        # Determine the id of the outputfile
        engine = self.inputs.config.create_engine()
        metadata_obj = sqla.MetaData()
        outputfile = sqla.Table("outputfile", metadata_obj, autoload_with=engine)

        idnum_query = sqla.select(outputfile.columns.idnum).where(
            outputfile.columns.filename == fname)
        with engine.commnect() as con:
            idnum_= con.execute(idnum_query).first()
        idnum = idnum_.idnum

        # Insert the image into the database
        engine = self.inputs.config.create_engine()
        metadata_obj = sqla.MetaData()
        table = sqla.Table("modelimages", metadata_obj, autoload_with=engine)
        
        insert_stmt = pg.insert(table).values(
            out_idnum = idnum,
            quantity = self.quantity,
            origin = self.origin.object,
            dims = self.dims,
            center = [c.value for c in self.center],
            width = [w.value for w in self.width],
            subobslongitude = self.subobslongitude.value,
            subobslatitude = self.subobslatitude.value,
            mechanism = self.mechanism,
            wavelength = [w.value for w in self.wavelength])

        with engine.connect() as con:
            result = con.execute(insert_stmt)
            con.commit()

        self.idnum = result.inserted_primary_key[0]
        savefile = os.path.join(os.path.dirname(fname), f'image.{self.idnum}.pkl')
        update = sqla.update(table).where(table.columns.idnum == self.idnum).values(
            filename=savefile)
        with engine.connect() as con:
            con.execute(update)
            con.commit()

        with open(savefile, 'wb') as f:
            pickle.dump((image, packets), f)

    def restore(self, fname, overwrite=False):
        # Determine the id of the outputfile
        engine = self.inputs.config.create_engine()
        metadata_obj = sqla.MetaData()
        outputfile = sqla.Table("outputfile", metadata_obj, autoload_with=engine)
        
        idnum_query = sqla.select(outputfile.columns.idnum).where(
            outputfile.columns.filename == fname)
        with engine.commnect() as con:
            idnum_= con.execute(idnum_query).first()
        oid = idnum_.idnum

        images = sqla.Table("modelimages", metadata_obj, autoload_with=engine)
        im_query = sqla.select(images.columns.filename).where(
            images.columns.out_idnum == oid,
            images.columns.quantity == self.quantity,
            images.columns.origin == self.origin.object,
            images.columns.dims == self.dims,
            images.columns.center == [s.value for s in self.center],
            images.columns.width == [w.value for w in self.width],
            images.columns.subobslongitude == self.subobslongitude.value,
            images.columns.subobslatitude == self.subobslatitude.value,
            images.columns.mechanism == self.mechanism,
            images.columns.wavelength == self.wavelength)
        
        with engine.connect() as con:
            result = con.execute(im_query).first()

        if (len(result) == 1) and overwrite:
            delete = sqla.delete(images).where(
                images.columns.filename == result.filename)
            if os.path.exists(result.filename):
                os.remove(result.filename)
            image, packets = None, None
        elif len(result) == 1:
            image, packets = pickle.load(open(result.filename[0], 'rb'))
        elif len(result) == 0:
            image, packets = None, None
        else:
            raise RuntimeError('ModelImage.restore',
                               'Should not be able to get here.')

        return image, packets

    def create_image(self, output):
        # Determine the proper frame rotation
        M = self.image_rotation()

        # Load data in solar reference frame
        packets = output.X
        if self.origin != self.inputs.geometry.planet:
            super().transform_reference_frame(packets)
        
        packets['radvel_sun'] = (packets['vy'] +
                                 output.vrplanet.to(self.unit/u.s).value)

        # packet positions in an array
        pts_sun = packets[['x', 'y', 'z']].values

        # Rotate to observer frame
        pts_obs = np.array(np.matmul(M, pts_sun.transpose()).transpose())

        # Determine which packets are not blocked by planet
        rhosqr_obs = np.linalg.norm(pts_obs[:, [0, 2]], axis=1)
        inview = (rhosqr_obs > 1) | (pts_obs[:,1] < 0)
        packets['frac'] *= inview

        # Which packets are in sunlight
        rhosqr_sun = np.linalg.norm(pts_sun[:, [0, 2]], axis=1)
        out_of_shadow = (rhosqr_sun > 1) | (pts_sun[:,1] < 0)

        # Packet weighting
        self.packet_weighting(packets, out_of_shadow, output.aplanet)
        packets['weight'] /= self.Apix

        pts_obs = pts_obs.transpose()
        range = [[x.value for x in self.xrange],
                 [z.value for z in self.zrange]]
        image = Histogram2d(pts_obs[0,:], pts_obs[2,:], weights=packets['weight'],
                            bins=self.dims, range=range)
        packim = Histogram2d(pts_obs[0,:], pts_obs[2,:], bins=self.dims, range=range)
        self.xaxis = image.x * self.unit
        self.zaxis = image.y * self.unit
        self.save(output.filename, image, packim)

        return image, packim

    def display(self, savefile='image.html', limits=None, show=True, log=True):
        if self.unit.__str__() == 'R_Mercury':
            ustr = 'R_M'
        else:
            ustr = 'R_obj'
            
        if self.quantity == 'radiance':
            runit = 'kR'
            rname = 'Radiance'
        elif self.quantity == 'column':
            runit = 'cm-2'
            rname = 'Column'
        else:
            assert 0

        tooltips = [('x', '$x{0.1f} ' + ustr),
                    ('y', '$y{0.1f} ' + ustr),
                    (rname, '@image ' + runit)]
        
        curdoc().theme = Theme(os.path.join(os.path.dirname(__file__),
                                            'data', 'bokeh.yml'))

        if log:
            if limits is None:
                limits = (self.image[self.image > 0].min(), self.image.max())
            else:
                pass
            color_mapper = LogColorMapper(palette=Inferno256, low=limits[0],
                                          high=limits[1])
        else:
            if limits is None:
                limits = (0, self.image.max())
            else:
                pass
            color_mapper = LinearColorMapper(palette=Inferno256, low=limits[0],
                                             high=limits[1])

        x0 = np.min(self.xaxis.value)
        y0 = np.min(self.zaxis.value)
        dw = np.max(self.xaxis.value) - np.min(self.xaxis.value)
        dh = np.max(self.zaxis.value) - np.min(self.zaxis.value)

        fig = bkp.figure(plot_width=1000, plot_height=1000,
                         title=f'{self.inputs.options.species} {rname}',
                         x_axis_label=f'Distance ({ustr})',
                         y_axis_label=f'Distance ({ustr})',
                         x_range=[np.min(self.xaxis.value),
                                  np.max(self.xaxis.value)],
                         y_range=[np.min(self.zaxis.value),
                                  np.max(self.zaxis.value)],
                         tooltips=tooltips)

        fig.image(image=[self.image.T], x=x0, y=y0, dw=dw, dh=dh,
                  color_mapper=color_mapper)
        xc = np.cos(np.linspace(0, 2*np.pi, 1000))
        yc = np.sin(np.linspace(0, 2*np.pi, 1000))
        fig.patch(xc, yc, fill_color='yellow')
        
        bkp.output_file(savefile)
        export_png(fig, filename=savefile.replace('.html', '.png'))
        if show:
            bkp.show(fig)
        else:
            bkp.save(fig)
            
        #
        # # Determine limits if none given
        # if limits is None:
        #     interval = PercentileInterval(95)
        #     self.blimits = interval.get_limits(self.image[self.image > 0])
        # elif len(limits) == 2:
        #     self.blimits = limits
        # else:
        #     assert 0, 'Problem with the display limits'

#        norm = ImageNormalize(self.image, stretch=LogStretch(),
#                              vmin=self.blimits[0], vmax=self.blimits[1])

        # # Make the colorbar
        # if self.quantity == 'column':
        #     clabel = f'$N_{{ {self.inputs.options.species} }}\ cm^{{-2}}$'
        # else:
        #     clabel = f'$I_{{ {self.inputs.options.species} }} R$'
        # cbar = fig.colorbar(im, shrink=0.7, label=clabel)

        # Put Planet's disk in the middle
        # xc, yc = (np.cos(np.linspace(0, 2*np.pi, 1000)),
        #           np.sin(np.linspace(0, 2*np.pi, 1000)))
        # ax.fill(xc, yc, 'y')

        return fig
    
    def image_rotation(self):
        slong = self.subobslongitude
        slat = self.subobslatitude
        
        pSun = np.array([0., -1., 0.])
        pObs = np.array([np.sin(slong)*np.cos(slat),
                         -np.cos(slong)*np.cos(slat),
                         np.sin(slat)])
        if np.array_equal(pSun, pObs):
            M = np.eye(3)
        else:
            costh = np.dot(pSun, pObs)/np.linalg.norm(pSun)/np.linalg.norm(pObs)
            theta = np.arccos(np.clip(costh, -1, 1))
            axis = np.cross(pSun, pObs)
            M = rotation_matrix(theta, axis)
        
        #M = np.transpose(M)
        return M

    def export(self, filename='image.json'):
        if filename.endswith('.json'):
            saveimage = {'image':self.image.tolist(),
                         'xaxis':self.xaxis.value.tolist(),
                         'zaxis':self.zaxis.value.tolist()}
            with open(filename, 'w') as f:
                json.dump(saveimage, f)
        else:
            raise TypeError('Not an valid file format')



'''Version of LOS that saves infomration in dataframes but won't work if too
many spectra'''
import os.path
import numpy as np
import pandas as pd
import pickle
import random
import copy
import astropy.units as u
from sklearn.neighbors import KDTree, BallTree
from scipy.interpolate import RectBivariateSpline
import scipy.linalg as lin
from mathMB import fit_model

from .ModelResults import ModelResult
from .database_connect import database_connect
from .Input import Input
from .Output import Output


xcols = ['x', 'y', 'z']
borecols = ['xbore', 'ybore', 'zbore']
NLONBINS, NLATBINS, NVELBINS = 72, 36, 100


class DataPackets:
    def __init__(self, *args, **kwargs):
        self.data = pd.DataFrame(*args, columns=['outputfile', 'specind',
                                                 'oind', 'weight', 'frac0',
                                                 'Index0'], *kwargs)
        self.totalsource = 0.
        self.atoms_per_packet = 0.
        self.outputfile = None
        self.out_idnum = None
        

class InputError(Exception):
    """Raised when a required parameter is not included."""
    def __init__(self, expression, message):
        self.expression = expression
        self.message = message


class LOSResult(ModelResult):
    """Class to contain the LOS result from multiple outputfiles."""
    def __init__(self, scdata, quantity='radiance', dphi=1*u.deg):
        """Determine column or emission along lines of sight.
        This assumes the model has already been run.
        
        Parameters
        ==========
        scdata
            Spacecraft data object (currently designed for MESSENGERdata object
            but can be faked for other types of data)
            
        quantity
            Quantity to calculate: 'column', 'radiance', 'density'
            
        dphi
            Angular size of the view cone. Default = r deg.
        """
        format_ = {'quantity':quantity}
        super().__init__(format_, species=scdata.species)
        
        # Basic information
        self.scdata = scdata
        self.scdata.set_frame('Model')
        self.type = 'LineOfSight'
        self.unit = None
        self.dphi = dphi.to(u.rad).value
        self.oedge = None
        
        self.fitted = None
        self.outputfiles = []
        nspec = len(self.scdata)
        self.radiance = pd.Series(np.zeros(nspec), index=scdata.data.index)
        self.npackets = pd.Series(np.zeros(nspec), index=scdata.data.index,
                                  dtype=int)
        self.totalsource = 0.
        self.sourcemap = None
        self.modelfiles = {}
        self.data_packets = DataPackets()

    def delete_models(self):
        """Deletes any LOSResult models associated with this data and input
        This may never actually do anything. Overwrite=True will also
        erase the outputfiles (which erases any models that depend on them).
        Unless I put separate outputfile and modelfile delete switches,
        This shouldn't do anything"""
        
        search_results = self.search()
        if len(search_results) != 0:
            print('Warning: LOSResult.delete_models found something to delete')
            for _, search_result in search_results.items():
                if search_result is not None:
                    idnum, modelfile = search_result
                    with database_connect() as con:
                        cur = con.cursor()
                        cur.execute(f'''DELETE from uvvsmodels
                                       WHERE idnum = %s''', (idnum, ))
                    if os.path.exists(modelfile):
                        os.remove(modelfile)
        else:
            pass

    def save(self, data_packets):
        # Insert the model into the database
        # Save is on an outputfile basis
        if self.quantity == 'radiance':
            mech = ', '.join(sorted([m for m in self.mechanism]))
            wave_ = sorted([w.value for w in self.wavelength])
            wave = ', '.join([str(w) for w in wave_])
        else:
            mech = None
            wave = None

        # Save query with all white space removed and lowercase
        tempname = f'temp_{str(random.randint(0, 1000000))}'

        with database_connect() as con:
            cur = con.cursor()
            cur.execute(f'''INSERT into uvvsmodels (out_idnum, quantity,
                            query, dphi, mechanism, wavelength,
                            fitted, filename)
                            values (%s, %s, %s, %s, %s, %s, %s, %s)''',
                        (data_packets.out_idnum, self.quantity,
                         self.scdata.query, self.dphi, mech, wave, self.fitted,
                         tempname))

            # Determine the savefile name
            idnum_ = pd.read_sql(f'''SELECT idnum
                                     FROM uvvsmodels
                                     WHERE filename='{tempname}';''', con)
            assert len(idnum_) == 1
            idnum = int(idnum_.idnum[0])

            savefile = os.path.join(os.path.dirname(data_packets.outputfile),
                                    f'model.{idnum}.pkl')

            cur.execute(f'''UPDATE uvvsmodels
                            SET filename=%s
                            WHERE idnum=%s''', (savefile, idnum))
            
        with open(savefile, 'wb') as f:
            pickle.dump(data_packets, f)
            
        return savefile

    def search(self):
        """
        :return: dictionary containing search results:
                 {outputfilename: (modelfile_id, modelfile_name)}
        """
        search_results = {}
        for outputfile in self.outputfiles:
            with database_connect() as con:
                # Determine the id of the outputfile
                idnum_ = pd.read_sql(
                    f'''SELECT idnum
                        FROM outputfile
                        WHERE filename='{outputfile}' ''', con)
                oid = idnum_.idnum[0]
            
                if self.quantity == 'radiance':
                    mech = ("mechanism = '" +
                            ", ".join(sorted([m for m in self.mechanism])) +
                            "'")
                    wave_ = sorted([w.value for w in self.wavelength])
                    wave = ("wavelength = '" +
                            ", ".join([str(w) for w in wave_]) +
                            "'")
                else:
                    mech = 'mechanism is NULL'
                    wave = 'wavelength is NULL'
            
                result = pd.read_sql(
                    f'''SELECT idnum, filename FROM uvvsmodels
                        WHERE out_idnum={oid} and
                              quantity = '{self.quantity}' and
                              query = '{self.scdata.query}' and
                              dphi = {self.dphi} and
                              {mech} and
                              {wave} and
                              fitted = {self.fitted}''', con)
            
                # Should only have one match per outputfile
                assert len(result) <= 1
                
                if len(result) == 0:
                    search_results[outputfile] = None
                else:
                    search_results[outputfile] = (result.iloc[0, 0],
                                                  result.iloc[0, 1])
                
        return search_results
    
    @staticmethod
    def restore(modelfile):
        # Restore is on an outputfile basis
        with open(modelfile, 'rb') as f:
            data_packets = pickle.load(f)

        return data_packets
    
    def _data_setup(self):
        # distance of s/c from planet
        data = self.scdata.data
        dist_from_plan = np.sqrt(data.x**2 + data.y**2 + data.z**2)
    
        # Angle between look direction and planet.
        ang = np.arccos((-data.x*data.xbore - data.y*data.ybore -
                         data.z*data.zbore)/dist_from_plan)
        
        # Check to see if look direction intersects the planet anywhere
        asize_plan = np.arcsin(1./dist_from_plan)

        # Don't worry about lines of sight that don't hit the planet
        dist_from_plan.loc[ang > asize_plan] = 1e30
        
        return dist_from_plan
    
    def _spectrum_process(self, spectrum, packets, tree, dist, i, ofile):
        x_sc = spectrum[xcols].values.astype(float)
        bore = spectrum[borecols].values.astype(float)
    
        dd = 30  # Furthest distance we need to look
        x_far = x_sc + bore * dd
        while np.linalg.norm(x_far) > self.oedge:
            dd -= 0.1
            x_far = x_sc + bore * dd
    
        t = [0.05]
        while t[-1] < dd:
            t.append(t[-1] + t[-1] * np.sin(self.dphi))
        t = np.array(t)
        Xbore = x_sc[np.newaxis, :] + bore[np.newaxis, :] * t[:, np.newaxis]
    
        wid = t * np.sin(self.dphi)
        ind = np.concatenate(tree.query_radius(Xbore, wid))
        ilocs = np.unique(ind).astype(int)
        indicies = packets.iloc[ilocs].index
        subset = packets.loc[indicies]
    
        xpr = subset[xcols] - x_sc[np.newaxis, :]
        rpr = np.sqrt(xpr['x'] * xpr['x'] +
                      xpr['y'] * xpr['y'] +
                      xpr['z'] * xpr['z'])
    
        losrad = np.sum(xpr * bore[np.newaxis, :], axis=1)
        inview = rpr < dist
        
        if np.any(inview):
            used_packets = subset[inview].index.to_list()
            used_packets0 = subset.loc[inview, 'Index']

            Apix = np.pi * (rpr[inview] * np.sin(self.dphi))**2 * (
                self.unit.to(u.cm))**2
            wtemp = subset.loc[inview, 'weight'] / Apix
            if self.quantity == 'radiance':
                # Determine if any packets are in shadow
                # Projection of packet onto LOS
                # Point along LOS the packet represents
                losrad_ = losrad[inview].values
                hit = (x_sc[np.newaxis, :] +
                       bore[np.newaxis, :] * losrad_[:, np.newaxis])
                rhohit = np.linalg.norm(hit[:, [0, 2]], axis=1)
                out_of_shadow = (rhohit > 1) | (hit[:, 1] < 0)
                wtemp *= out_of_shadow

                rad = wtemp.sum()
                # Save the weight information
                rat = spectrum.radiance/rad if rad > 0 else 0.
                processed = pd.DataFrame(
                    {'outputfile': [ofile for _ in used_packets],
                     'specind': [i for _ in used_packets],
                     'oind': used_packets,
                     'weight': wtemp.values,
                     'frac0': packets.loc[used_packets, 'frac0'].values,
                     'Index0': used_packets0.values,
                     'ratio': [rat for _ in used_packets]})
            else:
                assert False, 'Other quantities not set up.'
        else:
            processed = None
        
        return processed

    @staticmethod
    def _tree(values, treetype='KDTree'):
        if treetype == 'KDTree':
            return KDTree(values)
        elif treetype == 'BallTree':
            return BallTree(values)
        
    def determine_source_from_data(self, modnum, weight_method='scaling'):
        modkey = f'model{modnum:02d}'
        maskkey = f'mask{modnum:02d}'
        mask = self.scdata.data[maskkey]
        
        self.modelfiles = self.scdata.model_info[modkey]['modelfiles']
        assert len(self.modelfiles) > 0
        self.outputfiles = self.modelfiles.keys()
        surface_source = pd.DataFrame(columns=['outputfile', 'Index', 'longitude',
                                               'latitude', 'velocity'])
        data = self.scdata.data

        self.data_packets = DataPackets()
        nsteps, endtime = 0, 0
        for outputfile, modefile in self.modelfiles.items():
            # Restore the unfit output file
            output = Output.restore(outputfile)
            self.totalsource += output.totalsource
            endtime = output.inputs.options.endtime
            nsteps = output.nsteps
            data_packets_it = self.restore(modefile)
            
            self.data_packets.data = self.data_packets.data.append(
                data_packets_it.data)
            self.data_packets.totalsource += data_packets_it.totalsource
            
            vel_ = np.sqrt(output.X0.vx**2 + output.X0.vy**2 +
                           output.X0.vz**2) * output.inputs.geometry.planet.radius
            ind0 = data_packets_it.data.Index0.unique()
            surface_source = surface_source.append(pd.DataFrame(
                {'outputfile': [outputfile for _ in range(len(ind0))],
                 'Index': ind0,
                 'longitude': output.X0.loc[ind0, 'longitude'].values,
                 'latitude': output.X0.loc[ind0, 'latitude'].values,
                 'velocity': vel_.loc[ind0]}), ignore_index=True)
            del output
            del data_packets_it
            
        if weight_method == 'scaling':
            # Determine the proper weightings
            spec_ind = self.data_packets.data.set_index('specind')
            spec_ind = spec_ind.loc[mask]
            scale_factor_ = spec_ind.groupby(['outputfile', 'Index0'])['ratio'].mean()
            scale_factor_ /= scale_factor_[scale_factor_ > 0].mean()
    
            self.data_packets.data['scale_factor'] = np.zeros(len(self.data_packets.data))
            ind0_ind = self.data_packets.data.set_index(['outputfile', 'Index0'])
            ind0_ind.loc[scale_factor_.index, 'scale_factor'] = scale_factor_
            self.data_packets.data = ind0_ind.reset_index()
            self.data_packets.data['fit_weight'] = (self.data_packets.data.weight *
                                                    self.data_packets.data.scale_factor)
    
            # Determine surface weighting
            surface_ind = surface_source.set_index(['outputfile', 'Index'])
            surface_ind.loc[scale_factor_.index, 'weight'] = scale_factor_
            surface_ind.dropna(inplace=True)
            surface_source = surface_ind.reset_index()
        # elif weight_method == 'leastsq':
        #     # This method produces too many scale_factors < 0
        #     assert 0
        #
        #     # Solve linear equations to get best fit
        #     temp = self.data_packets.data.set_index('specind')
        #     temp = temp.loc[mask]
        #     temp.reset_index(inplace=True)
        #
        #     W = np.zeros((len(data), len(self.modelfiles),
        #                   self.data_packets.data.Index0.max()+1))
        #
        #     spec_ind = [data.index.get_loc(i) for i in temp.specind.values]
        #     pack_ind = temp.index.to_list()
        #
        #     outputfiles = {ofile: i for i, ofile
        #                    in enumerate(self.modelfiles.keys())}
        #     oo = [outputfiles[ofile] for ofile in temp.outputfile]
        #     W[spec_ind, oo, temp.Index0.to_list()] = temp.weight
        #
        #     W2 = W.squeeze(axis=1)/W.mean()
        #     # assert np.all(W == W2.reshape(W.shape))
        #     assert W.shape[1] == 1
        #     scale_factor_, r, rank, s = lin.lstsq(W2, data.loc[:, 'radiance'].values)
        #     scale_factor_ /= scale_factor_[scale_factor_ > 0].mean()
        #     self.data_packets.data['scale_factor'] = (
        #         scale_factor_[self.data_packets.data.Index0.to_list()])
        #     self.data_packets.data['fit_weight'] = (self.data_packets.data.weight *
        #                                             self.data_packets.data.scale_factor)
        #
        #     surface_source['weight'] = scale_factor_[surface_source.Index.to_list()]
        #
        #     scaled = pd.read_pickle('surface_source_scaled.pkl')
        #
        #     from IPython import embed; embed()
        #     import sys; sys.exit()
        #
        #
        #     self.data_packets.data['scale_factor'] = (
        #         scale_factor_[self.data_packets.data['Index0'].to_list()])
        #
        #     # Determine surface weighting
        #
        #     surface_source.dropna(inplace=True)
        #     # surface_source = surface_source[surface_source.weight > 0]
        else:
            raise InputError('LOSResult.determine_source_from_data: '
                             'Not a valid weighting method')

        initial = set(zip(self.data_packets.data.outputfile.to_list(),
                          self.data_packets.data.Index0.to_list(),
                          self.data_packets.data.frac0.to_list(),
                          self.data_packets.data.scale_factor.to_list()))
        frac0 = sum(x[2] * x[3] for x in initial)
        self.totalsource = frac0 * nsteps
        mod_rate = self.totalsource / endtime.value
        atoms_per_packet = 1e23 / mod_rate

        spec_group = self.data_packets.data.groupby('specind')
        fitrad_ = spec_group['fit_weight'].sum() * atoms_per_packet

        self.radiance.loc[fitrad_.index] = fitrad_
        self.radiance *= u.R

        # Create a new sourcemap
        source, xx, yy = np.histogram2d(surface_source['longitude'],
                                        surface_source['latitude'],
                                        weights=surface_source['weight'],
                                        range=[[0, 2*np.pi], [-np.pi/2, np.pi/2]],
                                        bins=(NLONBINS, NLATBINS))
        source = source/np.cos(yy+(yy[1]-yy[0])/2.)[np.newaxis,:-1]
        source[:,[0,-1]] = 0
        
        v_source, v = np.histogram(surface_source['velocity'], bins=NVELBINS,
                                   range=[0, surface_source['velocity'].max()],
                                   weights=surface_source['weight'])
        v_source /= np.max(v_source)

        # packets available
        packets, _, _ = np.histogram2d(surface_source['longitude'],
                                       surface_source['latitude'],
                                       range=[[0, 2*np.pi], [-np.pi/2, np.pi/2]],
                                       bins=(NLONBINS, NLATBINS))
        packets = packets/np.cos(yy+(yy[1]-yy[0])/2.)[np.newaxis,:-1]
        packets[:,[0,-1]] = 0
        v_packets, _ = np.histogram(surface_source['velocity'], bins=NVELBINS,
                                    range=[0, surface_source['velocity'].max()])
        v_packets = v_packets / np.max(v_packets)
        
        sourcemap = {'longitude': xx*u.rad,
                     'latitude': yy*u.rad,
                     'abundance': source,
                     'p_available': packets,
                     'velocity': v*u.km/u.s,
                     'vdist': v_source,
                     'v_available': v_packets,
                     'coordinate_system': 'solar-fixed'}
        self.sourcemap = sourcemap
        
    def simulate_data_from_inputs(self, inputs_, npackets, overwrite=False,
                                  packs_per_it=None):
        """Given a set of inputs, determine what the spacecraft should see.
        
        Parameters
        ==========
        inputs_
            A nexoclom Input object or the name of an inputs file
        """
        if isinstance(inputs_, str):
            self.inputs = Input(inputs_)
        elif isinstance(inputs_, Input):
            self.inputs = copy.deepcopy(inputs_)
        else:
            raise InputError('nexoclom.LOSResult', 'Problem with the inputs.')
    
        # TAA needs to match the data
        self.inputs.geometry.taa = self.scdata.taa
        self.unit = u.def_unit('R_' + self.inputs.geometry.planet.object,
                               self.inputs.geometry.planet.radius)
    
        # If using a planet-fixed source map, need to set subsolarlon
        if ((self.inputs.spatialdist.type == 'surface map') and
            (self.inputs.spatialdist.coordinate_system == 'planet-fixed')):
            self.inputs.spatialdist.subsolarlon = self.scdata.subslong.median() * u.rad
        else:
            pass
    
        # Run the model
        self.fitted = False
        self.inputs.run(npackets, packs_per_it=packs_per_it, overwrite=overwrite)
        self.search_for_outputs()
        search_results = self.search()
    
        dist_from_plan = (self._data_setup()
                          if None in search_results.values()
                          else None)

        data = self.scdata.data
        for outputfile, search_result in search_results.items():
            if search_result is None:
                # simulate the data
                output = Output.restore(outputfile)
                data_packets_it = DataPackets()
                data_packets_it.outputfile = outputfile
                data_packets_it.out_idnum = output.idnum
                data_packets_it.totalsource = output.totalsource

                packets = copy.deepcopy(output.X)
                packets['frac0'] = output.X0.loc[packets['Index'], 'frac'].values
                packets['radvel_sun'] = (packets['vy'] +
                                         output.vrplanet.to(self.unit / u.s).value)
                self.oedge = output.inputs.options.outeredge * 2
        
                # Will base shadow on line of sight, not the packets
                out_of_shadow = np.ones(len(packets))
                self.packet_weighting(packets, out_of_shadow, output.aplanet)
        
                # This sets limits on regions where packets might be
                tree = self._tree(packets[xcols].values)
        
                print(f'{data.shape[0]} spectra taken.')
                for i, spectrum in data.iterrows():
                    processed = self._spectrum_process(spectrum, packets, tree,
                                                       dist_from_plan[i],
                                                       i, outputfile)
                    if processed is not None:
                        data_packets_it.data = data_packets_it.data.append(
                            processed, ignore_index=True)
                    else:
                        pass
                    
                    ind = data.index.get_loc(i)
                    if (ind % (len(data)//10)) == 0:
                        print(f'Completed {ind+1} spectra')
                    else:
                        pass

                self.data_packets.data = self.data_packets.data.append(
                    data_packets_it.data)
                self.data_packets.totalsource += data_packets_it.totalsource
                
                modelfile = self.save(data_packets_it)
                self.modelfiles[outputfile] = modelfile
                
                del output
                del data_packets_it
            else:
                data_packets_it = self.restore(search_result[1])
                self.data_packets.data = self.data_packets.data.append(
                    data_packets_it.data)
                self.data_packets.totalsource += data_packets_it.totalsource
                self.modelfiles[outputfile] = search_result[1]
                del data_packets_it

        mod_rate = self.data_packets.totalsource / self.inputs.options.endtime.value
        atoms_per_packet = 1e23 / mod_rate
        
        spec_group = self.data_packets.data.groupby('specind')
        rad_ = spec_group['weight'].sum() * atoms_per_packet
        self.radiance.loc[rad_.index] = rad_
        self.radiance *= u.R

    def recompute_radiance(self, modnum, new_source):
        data = self.scdata.data
        radiance = pd.Series(index=data.index)

        modkey = f'model{modnum:02d}'
        maskkey = f'mask{modnum:02d}'
        
        self.modelfiles = self.scdata.model_info[modkey]['modelfiles']
        assert len(self.modelfiles) > 0
        self.outputfiles = self.modelfiles.keys()

        self.data_packets = DataPackets()
        surface_source = pd.DataFrame(columns=['outputfile', 'Index', 'longitude',
                                               'latitude'])
        nsteps, endtime = 0, 0
        for outputfile, modefile in self.modelfiles.items():
            # Restore the unfit output file
            output = Output.restore(outputfile)
            self.totalsource += output.totalsource
            endtime = output.inputs.options.endtime
            nsteps = output.nsteps
            data_packets_it = self.restore(modefile)
            
            self.data_packets.data = self.data_packets.data.append(
                data_packets_it.data)
            self.data_packets.totalsource += data_packets_it.totalsource
            
            ind0 = data_packets_it.data.Index0.unique()
            surface_source = surface_source.append(pd.DataFrame(
                {'outputfile': [outputfile for _ in range(len(ind0))],
                 'Index': ind0,
                 'longitude': output.X0.loc[ind0, 'longitude'].values,
                 'latitude': output.X0.loc[ind0, 'latitude'].values}), ignore_index=True)

            del output
            del data_packets_it

        # Determine new weighting from given map
        new_weight_fn = RectBivariateSpline(new_source['longitude'][:-1].value,
                                            np.sin(new_source['latitude'][:-1].value),
                                            new_source['abundance'])
        scale_factor_ = new_weight_fn(surface_source.longitude,
                                      np.sin(surface_source.latitude),
                                      grid=False)
        scale_factor_ /= scale_factor_[scale_factor_ > 0].mean()
        surface_source['scale_factor'] = scale_factor_
        
        surface_ind = surface_source.set_index(['outputfile', 'Index'])
        ind0_ind = self.data_packets.data.set_index(['outputfile', 'Index0'])
        ind0_ind.loc[surface_ind.index, 'scale_factor'] = surface_ind.scale_factor
        # ind0_ind = ind0_ind.dropna()
        
        self.data_packets.data = ind0_ind.reset_index()
        self.data_packets.data['new_weight'] = (self.data_packets.data.weight *
                                                self.data_packets.data.scale_factor)
        
        initial = set(zip(self.data_packets.data.outputfile.to_list(),
                          self.data_packets.data.Index0.to_list(),
                          self.data_packets.data.frac0.to_list(),
                          self.data_packets.data.scale_factor.to_list()))
        frac0 = sum(x[2] * x[3] for x in initial)
        self.totalsource = frac0 * nsteps
        mod_rate = self.totalsource / endtime.value
        atoms_per_packet = 1e23 / mod_rate

        spec_group = self.data_packets.data.groupby('specind')
        fitrad_ = spec_group['new_weight'].sum() * atoms_per_packet
        # fitrad = spec_group['weight'].sum() * atoms_per_packet

        self.radiance.loc[fitrad_.index] = fitrad_
        self.radiance *= u.R


if __name__ == '__main__':
    import matplotlib.pyplot as plt
    hist, xx, yy = np.histogram2d(surface_ind.longitude, surface_ind.latitude,
                                  weights=surface_ind.scale_factor,
                                  range=[[0, 2*np.pi], [-np.pi/2, np.pi/2]],
                                  bins=(NLONBINS, NLATBINS))
    hist = hist / np.cos(yy + (yy[1] - yy[0]) / 2.)[np.newaxis, :-1]
    hist *= new_source['abundance'].mean()/hist.mean()

    fig, ax = plt.subplots(1, 2)
    ax[0].imshow(new_source['abundance'], vmin=0, vmax=100)
    ax[1].imshow(hist, vmin=0, vmax=100)
    plt.show()
    #

    plt.plot(np.sum(new_source['abundance'], axis=0))
    plt.plot(np.sum(hist, axis=0))
    plt.show()
    plt.plot(np.sum(new_source['abundance'], axis=1))
    plt.plot(np.sum(hist, axis=1))
    plt.show()
    #
    # from IPython import embed; embed()
    # import sys; sys.exit()



import os
import numpy as np
import pandas as pd
import astropy.units as u
from nexoclom.solarsystem import SSObject, planet_dist
import pytest


table = pd.read_csv(os.path.join(os.path.dirname(__file__),
                                 'perihelion_aphelion.csv'))

planet = table.planet.values
perihelion = table.perihelion.apply(lambda x: (x*u.km).to(u.au).value).values
aphelion = table.aphelion.apply(lambda x: (x*u.km).to(u.au).value).values

ind = [0, 5, 7]
args_dist = [(planet[ind[0]], 0, (perihelion[ind[0]], 0)),
             (planet[ind[1]], np.pi*u.rad, (aphelion[ind[1]], 0)),
             (SSObject(planet[ind[2]]), 0*u.deg, (perihelion[ind[2]], 0))]

@pytest.mark.solarsystem
@pytest.mark.parametrize('planet, taa, result', args_dist)
def test_planet_dist(planet, taa, result):
    r, v_r = planet_dist(planet, taa)
    
    assert r.value == pytest.approx(result[0], rel=0.01)
    assert v_r.value == pytest.approx(result[1], rel=0.01)
    assert (r.unit, v_r.unit) == (u.au, u.km/u.s)

@pytest.mark.solarsystem
def test_not_real_object():
    assert planet_dist('Fake') is None
    
@pytest.mark.solarsystem
def test_planet_dist_bad_input():
    with pytest.raises(TypeError):
        planet_dist(4)

    with pytest.raises(TypeError):
        planet_dist('Jupiter', 'bad')

@pytest.mark.solarsystem
@pytest.mark.xfail
def test_planet_dist_with_time():
    # Not setup to work with times yet
    r, v_r = planet_dist('Mercury', time='2021-10-20T12:00:00')
    assert r.value == pytest.approx(4, rel=0.01)
    assert v_r.value == pytest.approx(4, rel=0.01)
    assert (r.unit, v_r.unit) == (u.au, u.km/u.s)

@pytest.mark.solarsystem
def test_planet_dist_no_inputs():
    assert planet_dist('Mercury') is None
    


import pytest
import os
import numpy as np
import pandas as pd
import io
from nexoclom import __file__ as basefile
from nexoclom.atomicdata import gValue, PhotoRate
from nexoclom.atomicdata.initialize_atomicdata import (make_gvalue_table,
                                                       make_photorates_table)


@pytest.mark.atomicdata
def test_make_gvalue_table(monkeypatch, tmpdir):
    gvalue_true_file = os.path.join(os.path.dirname(basefile), 'data', 'g-values',
                                    'g-values.pkl')
    gvalue_test_file = tmpdir.join('g-values_test.pkl')

    monkeypatch.setattr(gValue, "gvalue_filename", lambda: gvalue_test_file)
    # monkeypatch.setattr(gValue, 'gvalue_file', gvalue_test_file)
    make_gvalue_table()

    gvalue_true = pd.read_pickle(gvalue_true_file)
    gvalue_test = pd.read_pickle(gvalue_test_file)
    assert np.all(gvalue_true == gvalue_test)


@pytest.mark.atomicdata
def test_make_photorates_table(monkeypatch, tmpdir):
    monkeypatch.setattr('sys.stdin', io.StringIO('0'))
    photo_true_file = os.path.join(os.path.dirname(basefile), 'data', 'Loss',
                                   'photorates.pkl')
    photo_test_file = tmpdir.join('photorates.pkl')

    monkeypatch.setattr(PhotoRate, 'photorates_filename', lambda: photo_test_file)
    make_photorates_table()
 
    photo_true = pd.read_pickle(photo_true_file)
    photo_test = pd.read_pickle(photo_test_file)
    assert np.all(photo_true == photo_test)



from nexoclom.utilities import NexoclomConfig
import os
import shutil
import subprocess
import pytest


configfiles = ['.nexoclom', '.nexoclom_test', '.nexoclom_test2']
results = [{'savepath':'/Users/mburger/Work/Research/modeloutputs',
            'database':'thesolarsystemmb',
            'port':5432,
            'dbhost': False},
           {'savepath':'/Users/mburger/Work/Research/modeloutputs_test',
            'database':'thesolarsystemmb_test',
            'port':5432,
            'dbhost': False},
           {'savepath':'/Users/mburger/Work/Research/modeloutputs2',
            'database':'thesolarsystemmb',
            'port':1234,
            'dbhost': 'dlhlalab1.stsci.edu'}]


@pytest.mark.utilities
@pytest.mark.parametrize('inputs', zip(configfiles, results))
def test_read_configfile_and_database_connect(inputs):
    def compare_config(conf, res):
        assert conf.savepath == res['savepath']
        assert conf.database == res['database']
        assert conf.port == res['port']
        assert conf.dbhost == res['dbhost']
        return 0
        
    def compare_database(con, res):
        assert con.autocommit
        assert con.info.dbname == res['database']
        return 0
    
    configfile, result = inputs
    os.environ['NEXOCLOMCONFIG'] = os.path.join(os.environ['HOME'], configfile)
    if configfile == '.nexoclom_test':
        # Test with environment variable
        # config file = '/Users/mburger/.nexoclom_test'
        if os.path.exists(result['savepath']):
            shutil.rmtree(result['savepath'])
        else:
            pass
        
        # Verify NexoclomConfig() works
        config = NexoclomConfig()
        assert os.path.exists(result['savepath'])
        compare_config(config, result)
        
        # Verify config.verify_database_running() works
        subprocess.run('pg_ctl stop', capture_output=True, shell=True)
        # verify database starts up
        assert config.verify_database_running() == 'Started Database'

        # verify database running
        assert config.verify_database_running() == 'Database Already Running'
        
        # Verify config.database_connect works
        compare_database(config.database_connect(), result)
    elif configfile == '.nexoclom':
        del os.environ['NEXOCLOMCONFIG']
        
        config = NexoclomConfig()
        compare_config(config, result)
    elif configfile == '.nexoclom_test2':
        # Test with different port and external database
        config = NexoclomConfig(os.path.join(os.environ['HOME'], configfile))
        compare_config(config, result)
if __name__ == '__main__':
    for param in zip(configfiles, results):
        test_read_configfile_and_database_connect(param)



from nexoclom.solarsystem.SSObject import SSObject
from nexoclom.solarsystem.planet_dist import planet_dist
from nexoclom.solarsystem.planet_geometry import planet_geometry

__name__ = 'solarsystem'
__author__ = 'Matthew Burger'
__email__ = 'mburger@stsci.edu'
